# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import pytest
import biotite.sequence.align as align
import biotite.application.muscle as muscle
from biotite.application import VersionError
from ...util import is_not_installed
from .util import sequences



@pytest.mark.skipif(
    is_not_installed("muscle"),
    reason="MUSCLE is not installed"
)
@pytest.mark.parametrize("gap_penalty", [-10, (-10,-1)])
def test_align_multiple(sequences, gap_penalty):
    r"""
    Test `align_multiple()` function using actual long sequences,
    compared to the output of MUSCLE.
    Both alignment methods are heuristic, the exact same result is not
    expected.
    Just assert that the resulting score is at least the 50 % of the
    score of the MUSCLE alignment.
    """
    matrix = align.SubstitutionMatrix.std_protein_matrix()
    
    test_alignment, order, tree, distances = align.align_multiple(
        sequences, matrix, gap_penalty=gap_penalty, terminal_penalty=True
    )
    test_score = align.score(
        test_alignment, matrix, gap_penalty, terminal_penalty=True
    )
    
    try:
        ref_alignment = muscle.MuscleApp.align(
            sequences, matrix=matrix, gap_penalty=gap_penalty
        )
    except VersionError:
        pytest.skip(f"Invalid Muscle software version")
    ref_score = align.score(
        ref_alignment, matrix, gap_penalty, terminal_penalty=True
    )
    
    assert test_score >= ref_score * 0.5


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.sequence"
__author__ = "Patrick Kunzmann"
__all__ = ["Alphabet", "LetterAlphabet", "AlphabetMapper", "AlphabetError"]

import copy
from numbers import Integral
import string
import numpy as np
from .codec import encode_chars, decode_to_chars, map_sequence_code


class Alphabet(object):
    """
    This class defines the allowed symbols for a :class:`Sequence` and
    handles the encoding/decoding between symbols and symbol codes.
    
    An :class:`Alphabet` is created with the list of symbols, that can
    be used in this context.
    In most cases a symbol will be simply a letter, hence a string of
    length 1. But in principle every hashable Python object can serve
    as symbol.
    
    The encoding of a symbol into a symbol code is
    done in the following way: Find the first index in the symbol list,
    where the list element equals the symbol. This index is the
    symbol code. If the symbol is not found in the list, an
    :class:`AlphabetError` is raised.
    
    Internally, a dictionary is used for encoding, with symbols as keys
    and symbol codes as values. Therefore, every symbol must be
    hashable. For decoding the symbol list is indexed with the symbol
    code.
    
    If an alphabet *1* contains the same symbols and the same
    symbol-code-mappings like another alphabet *2*, but alphabet *1*
    introdues also new symbols, then alphabet *1* *extends* alphabet
    *2*.
    Per definition, every alphabet also extends itself.
    
    Objects of this class are immutable.
    
    Parameters
    ----------
    symbols : iterable object
        The symbols, that are allowed in this alphabet. The
        corresponding code for a symbol, is the index of that symbol
        in this list.
    
    Examples
    --------
    Create an Alphabet containing DNA letters and encode/decode a
    letter/code:
    
    >>> alph = Alphabet(["A","C","G","T"])
    >>> print(alph.encode("G"))
    2
    >>> print(alph.decode(2))
    G
    >>> try:
    ...    alph.encode("foo")
    ... except Exception as e:
    ...    print(e)
    Symbol 'foo' is not in the alphabet
    
    Create an Alphabet of arbitrary objects:
    
    >>> alph = Alphabet(["foo", 42, (1,2,3), 5, 3.141])
    >>> print(alph.encode((1,2,3)))
    2
    >>> print(alph.decode(4))
    3.141

    On the subject of alphabet extension:
    An alphabet always extends itself.
    
    >>> Alphabet(["A","C","G","T"]).extends(Alphabet(["A","C","G","T"]))
    True

    An alphabet extends an alphabet when it contains additional symbols...
    
    >>> Alphabet(["A","C","G","T","U"]).extends(Alphabet(["A","C","G","T"]))
    True
    
    ...but not vice versa
    
    >>> Alphabet(["A","C","G","T"]).extends(Alphabet(["A","C","G","T","U"]))
    False
    
    Two alphabets with same symbols but different symbol-code-mappings
    
    >>> Alphabet(["A","C","G","T"]).extends(Alphabet(["A","C","T","G"]))    
    False
    """
    
    def __init__(self, symbols):
        if len(symbols) == 0:
            raise ValueError("Symbol list is empty")
        self._symbols = copy.deepcopy(list(symbols))
        self._symbol_dict = {}
        for i, symbol in enumerate(symbols):
            self._symbol_dict[symbol] = i

    def __repr__(self):
        """Represent Alphabet as a string for debugging."""
        return f'Alphabet({self._symbols})'

    def get_symbols(self):
        """
        Get the symbols in the alphabet.
        
        Returns
        -------
        symbols : list
            Copy of the internal list of symbols.
        """
        return copy.deepcopy(self._symbols)
    
    def extends(self, alphabet):
        """
        Check, if this alphabet extends another alphabet.
        
        Parameters
        ----------
        alphabet : Alphabet
            The potential parent alphabet.

        Returns
        -------
        result : bool
            True, if this object extends `alphabet`, false otherwise.
        """
        if alphabet is self:
            return True
        elif len(alphabet) > len(self):
            return False
        else:
            return alphabet.get_symbols() \
                == self.get_symbols()[:len(alphabet)]
    
    def encode(self, symbol):
        """
        Use the alphabet to encode a symbol.
        
        Parameters
        ----------
        symbol : object
            The object to encode into a symbol code.
        
        Returns
        -------
        code : int
            The symbol code of `symbol`.
        
        Raises
        ------
        AlphabetError
            If `symbol` is not in the alphabet.
        """
        try:
            return self._symbol_dict[symbol]
        except KeyError:
            raise AlphabetError(
                f"Symbol {repr(symbol)} is not in the alphabet"
            )
    
    def decode(self, code):
        """
        Use the alphabet to decode a symbol code.
        
        Parameters
        ----------
        code : int
            The symbol code to be decoded.
        
        Returns
        -------
        symbol : object
            The symbol corresponding to `code`.
        
        Raises
        ------
        AlphabetError
            If `code` is not a valid code in the alphabet.
        """
        if code < 0 or code >= len(self._symbols):
            raise AlphabetError(f"'{code:d}' is not a valid code")
        return self._symbols[code]
    
    def encode_multiple(self, symbols, dtype=np.int64):
        """
        Encode a list of symbols.
        
        Parameters
        ----------
        symbols : array-like
            The symbols to encode.
        dtype : dtype, optional
            The dtype of the output ndarray. (Default: `int64`)
            
        Returns
        -------
        code : ndarray
            The sequence code.
        """
        return np.array([self.encode(e) for e in symbols], dtype=dtype)
    
    def decode_multiple(self, code):
        """
        Decode a sequence code into a list of symbols.
        
        Parameters
        ----------
        code : ndarray
            The sequence code to decode.
        
        Returns
        -------
        symbols : list
            The decoded list of symbols.
        """
        return [self.decode(c) for c in code]
    
    def is_letter_alphabet(self):
        """
        Check whether the symbols in this alphabet are single printable
        letters.
        If so, the alphabet could be expressed by a `LetterAlphabet`.

        Returns
        -------
        is_letter_alphabet : bool
            True, if all symbols in the alphabet are 'str' or 'bytes',
            have length 1 and are printable.
        """
        for symbol in self:
            if not isinstance(symbol, (str, bytes)) \
                or len(symbol) > 1:
                    return False
            if isinstance(symbol, str):
                symbol = symbol.encode("ASCII")
            if symbol not in LetterAlphabet.PRINATBLES:
                return False
        return True
    
    def __str__(self):
        return str(self.get_symbols())
    
    def __len__(self):
        return len(self.get_symbols())
    
    def __iter__(self):
        return self.get_symbols().__iter__()
    
    def __contains__(self, symbol):
        return symbol in self.get_symbols()
    
    def __hash__(self):
        return hash(tuple(self._symbols))
    
    def __eq__(self, item):
        if item is self:
            return True
        if not isinstance(item, Alphabet):
            return False
        return self.get_symbols() == item.get_symbols()


class LetterAlphabet(Alphabet):
    """
    :class:`LetterAlphabet` is a an :class:`Alphabet` subclass
    specialized for letter based alphabets, like DNA or protein
    sequence alphabets.
    The alphabet size is limited to the 94 printable, non-whitespace
    characters.
    Internally the symbols are saved as `bytes` objects.
    The encoding and decoding process is a lot faster than for a
    normal :class:`Alphabet`.

    The performance gain comes through the use of *NumPy* and *Cython*
    for encoding and decoding, without the need of a dictionary.

    Parameters
    ----------
    symbols : iterable object or str or bytes
        The symbols, that are allowed in this alphabet. The
        corresponding code for a symbol, is the index of that symbol
        in this list.
    """
    
    PRINATBLES = (string.digits + string.ascii_letters + string.punctuation) \
                 .encode("ASCII")

    def __init__(self, symbols):
        if len(symbols) == 0:
            raise ValueError("Symbol list is empty")
        self._symbols = []
        for symbol in symbols:
            if not isinstance(symbol, (str, bytes)) or len(symbol) > 1:
                raise ValueError(f"Symbol '{symbol}' is not a single letter")
            if isinstance(symbol, str):
                symbol = symbol.encode("ASCII")
            if symbol not in LetterAlphabet.PRINATBLES:
                raise ValueError(
                    f"Symbol {repr(symbol)} is not printable or whitespace"
                )
            self._symbols.append(symbol)
        # Direct 'astype' conversion is not allowed by numpy
        # -> frombuffer()
        self._symbols = np.frombuffer(
            np.array(self._symbols, dtype="|S1"),
            dtype=np.ubyte
        )

    def __repr__(self):
        """Represent LetterAlphabet as a string for debugging."""
        return f'LetterAlphabet({self.get_symbols()})'
    
    def extends(self, alphabet):
        if alphabet is self:
            return True
        elif type(alphabet) == LetterAlphabet:
            return np.all(
                alphabet._symbols == self._symbols[:len(alphabet._symbols)]
            )
        else:
            return super().extends(alphabet)

    def get_symbols(self):
        """
        Get the symbols in the alphabet.
        
        Returns
        -------
        symbols : list
            Copy of the internal list of symbols.
        """
        return [symbol.decode("ASCII") for symbol
                in self._symbols_as_bytes()]
    
    def encode(self, symbol):
        if not isinstance(symbol, (str, bytes)) or len(symbol) > 1:
            raise AlphabetError(f"Symbol '{symbol}' is not a single letter")
        indices = np.where(self._symbols == ord(symbol))[0]
        if len(indices) == 0:
            raise AlphabetError(
                f"Symbol {repr(symbol)} is not in the alphabet"
            )
        return indices[0]
    
    def decode(self, code, as_bytes=False):
        if code < 0 or code >= len(self._symbols):
            raise AlphabetError(f"'{code:d}' is not a valid code")
        return chr(self._symbols[code])
    
    def encode_multiple(self, symbols, dtype=None):
        """
        Encode multiple symbols.
        
        Parameters
        ----------
        symbols : iterable object or str or bytes
            The symbols to encode. The method is fastest when a
            :class:`ndarray`, :class:`str` or :class:`bytes` object
            containing the symbols is provided, instead of e.g. a list.
        dtype : dtype, optional
            For compatibility with superclass. The value is ignored
            
        Returns
        -------
        code : ndarray
            The sequence code.
        """
        if isinstance(symbols, str):
            symbols = np.frombuffer(symbols.encode("ASCII"), dtype=np.ubyte)
        elif isinstance(symbols, bytes):
            symbols = np.frombuffer(symbols, dtype=np.ubyte)
        elif isinstance(symbols, np.ndarray):
            symbols = np.frombuffer(
                symbols.astype(dtype="|S1"), dtype=np.ubyte
            )
        else:
            symbols = np.frombuffer(
                np.array(list(symbols), dtype="|S1"),
                dtype=np.ubyte
            )
        return encode_chars(alphabet=self._symbols, symbols=symbols)
    
    def decode_multiple(self, code, as_bytes=False):
        """
        Decode a sequence code into a list of symbols.
        
        Parameters
        ----------
        code : ndarray, dtype=uint8
            The sequence code to decode.
            Works fastest if a :class:`ndarray` is provided.
        as_bytes : bool, optional
            If true, the output array will contain `bytes`
            (dtype 'S1').
            Otherwise, the the output array will contain `str`
            (dtype 'U1').

        Returns
        -------
        symbols : ndarray, dtype='U1' or dtype='S1'
            The decoded list of symbols.
        """
        if not isinstance(code, np.ndarray):
            code = np.array(code, dtype=np.uint8)
        code = code.astype(np.uint8, copy=False)
        symbols = decode_to_chars(alphabet=self._symbols, code=code)
        # Symbols must be convverted from 'np.ubyte' to '|S1'
        symbols = np.frombuffer(symbols, dtype="|S1")
        if not as_bytes:
            symbols = symbols.astype("U1")
        return symbols
    
    def __contains__(self, symbol):
        if not isinstance(symbol, (str, bytes)):
            return False
        return ord(symbol) in self._symbols
    
    def __len__(self):
        return len(self._symbols)
    
    def _symbols_as_bytes(self):
        "Properly convert from dtype 'np.ubyte' to '|S1'"
        return np.frombuffer(self._symbols, dtype="|S1")

    

class AlphabetMapper(object):
    """
    This class is used for symbol code conversion from a source
    alphabet into a target alphabet.

    This means that the symbol codes are converted from one to another
    alphabet so that the symbol itself is preserved.
    This class works for single symbol codes or an entire sequence code
    likewise.
    
    Parameters
    ----------
    source_alphabet, target_alphabet : Alphabet
        The codes are converted from the source alphabet into the
        target alphabet.
        The target alphabet must contain at least all symbols of the
        source alphabet, but it is not required that the shared symbols
        are in the same order.
    
    Examples
    --------

    >>> source_alph = Alphabet(["A","C","G","T"])
    >>> target_alph = Alphabet(["T","U","A","G","C"])
    >>> mapper = AlphabetMapper(source_alph, target_alph)
    >>> print(mapper[0])
    2
    >>> print(mapper[1])
    4
    >>> print(mapper[[1,1,3]])
    [4 4 0]
    >>> in_sequence = GeneralSequence(source_alph, "GCCTAT")
    >>> print(in_sequence.code)
    [2 1 1 3 0 3]
    >>> print(in_sequence)
    GCCTAT
    >>> out_sequence = GeneralSequence(target_alph)
    >>> out_sequence.code = mapper[in_sequence.code]
    >>> print(out_sequence.code)
    [3 4 4 0 2 0]
    >>> print(out_sequence)
    GCCTAT
    """
    
    def __init__(self, source_alphabet, target_alphabet):
        if target_alphabet.extends(source_alphabet):
            self._necessary_mapping = False
        else:
            self._necessary_mapping = True
            self._mapper = np.zeros(
                len(source_alphabet),
                dtype=AlphabetMapper._dtype(len(target_alphabet))
            )
            for old_code in range(len(source_alphabet)):
                symbol = source_alphabet.decode(old_code)
                new_code = target_alphabet.encode(symbol)
                self._mapper[old_code] = new_code
        
    def __getitem__(self, code):
        if isinstance(code, Integral):
            if self._necessary_mapping:
                return self._mapper[code]
            else:
                return code
        if not isinstance(code, np.ndarray) \
           or code.dtype not in (np.uint8, np.uint16, np.uint32, np.uint64):
                code = np.array(code, dtype=np.uint64)
        if self._necessary_mapping:
            mapped_code = np.empty(len(code), dtype=self._mapper.dtype)
            map_sequence_code(
                self._mapper,
                code,
                mapped_code
            )
            return mapped_code
        else:
            return code


    @staticmethod
    def _dtype(alphabet_size):
        _size_uint8  = np.iinfo(np.uint8 ).max +1
        _size_uint16 = np.iinfo(np.uint16).max +1
        _size_uint32 = np.iinfo(np.uint32).max +1
        if alphabet_size <= _size_uint8:
            return np.uint8
        elif alphabet_size <= _size_uint16:
            return np.uint16
        elif alphabet_size <= _size_uint32:
            return np.uint32
        else:
            return np.uint64


class AlphabetError(Exception):
    """
    This exception is raised, when a code or a symbol is not in an
    :class:`Alphabet`.
    """
    pass


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite"
__author__ = "Patrick Kunzmann"
__all__ = ["colors", "set_font_size_in_coord"]

import abc
from collections import OrderedDict
import numpy as np
from numpy.linalg import norm


# Biotite themed colors
colors = OrderedDict([
    ("brightorange" , "#ffb569ff"),
    ("lightorange"  , "#ff982dff"),
    ("orange"       , "#ff8405ff"),
    ("dimorange"    , "#dc7000ff"),
    ("darkorange"   , "#b45c00ff"),
    ("brightgreen"  , "#98e97fff"),
    ("lightgreen"   , "#6fe04cff"),
    ("green"        , "#52da2aff"),
    ("dimgreen"     , "#45bc20ff"),
    ("darkgreen"    , "#389a1aff"),
])


def set_font_size_in_coord(text, width=None, height=None, mode="unlocked"):
    """
    Specifiy the font size of an existing `Text` object in coordinates
    of the object's reference coordiante system.

    Instead of having the font size fixed in 'pt', the size of the text
    scales to the specied width/height and adapts to changes in the
    plot's width/height.
    The scaling can be proportional or non-proportional, depending 
    the `mode`.

    Parameters
    ----------
    text : Text:
        The matplotlib `Text` to be scaled.
    width, height : float, optional
        The new width/height of `text` in its
        reference coordinate system.
        At least one value must be supplied.
    mode : {'proportional', 'unlocked', 'maximum', 'minimum'}, optional
        The scaling mode:

            - *proportional* - The width and height are scaled by the
              same extent.
              Either `width` or `height` must be set for this mode.
            - *unlocked* - The width and the height are scaled by
              different extents, changing the aspect ratio.
              Both `width` and `height` must be set for this mode.
            - *maximum* - The width and the height are scaled by
              the same extent, so that they are at maximum as large
              as the supplied `width`/`height`.
              Both `width` and `height` must be set for this mode.
            - *minimum* - The width and the height are scaled by
              the same extent, so that they are at minimum as large
              as the supplied `width`/`height`.
              Both `width` and `height` must be set for this mode.

    Notes
    -----
    This function uses the :func:`get_window_extent()` method of the
    :class:`Text` object.
    According to experience, this function does not give the the exact
    visual boundaries of the text.
    Consequently, the scaled text might be slightly smaller or larger
    than the specified width/height.
    This behavior is not equal for all initial font sizes (in 'pt'),
    the boundaries for an initial size of 1 'pt' seem to be most exact.
    """
    from matplotlib.transforms import Bbox
    from matplotlib.patheffects import AbstractPathEffect

    class TextScaler(AbstractPathEffect):
        def __init__(self, text, width, height, mode):
            self._text = text
            self._mode = mode
            self._width = width
            self._height = height

        def draw_path(self, renderer, gc, tpath, affine, rgbFace=None):
            ax = self._text.axes
            try:
                renderer = ax.get_figure().canvas.get_renderer()
            except:
                # Use cached renderer for backends, where
                # `get_renderer()` is not available
                # Based on the strategy from `Text.get_window_extent()`
                renderer = ax.get_figure()._cachedRenderer
            if renderer is None:
                raise
            bbox = text.get_window_extent(renderer)
            bbox = Bbox(ax.transData.inverted().transform(bbox))
            
            if self._mode == "proportional":
                if self._width is None:
                    # Proportional scaling based on height
                    scale_y = self._height / bbox.height
                    scale_x = scale_y
                elif self._height is None:
                    # Proportional scaling based on width
                    scale_x = self._width / bbox.width
                    scale_y = scale_x
            elif self._mode == "unlocked":
                scale_x = self._width / bbox.width
                scale_y = self._height / bbox.height
            elif self._mode == "minimum":
                scale_x = self._width / bbox.width
                scale_y = self._height / bbox.height
                scale = max(scale_x, scale_y)
                scale_x, scale_y = scale, scale
            elif self._mode == "maximum":
                scale_x = self._width / bbox.width
                scale_y = self._height / bbox.height
                scale = min(scale_x, scale_y)
                scale_x, scale_y = scale, scale

            affine = affine.identity().scale(scale_x, scale_y) + affine
            renderer.draw_path(gc, tpath, affine, rgbFace)
    
    if mode in ["unlocked", "minimum", "maximum"]:
        if width is None or height is None:
            raise TypeError(
                f"Width and height must be set in '{mode}' mode"
            )
    elif mode == "proportional":
        if  not (width  is None and height is not None) or \
            not (height is None and width  is not None):
                raise TypeError(
                    f"Either width or height must be set in '{mode}' mode"
                )
    else:
        raise ValueError(
                f"Unknown mode '{mode}'"
            )
    text.set_path_effects([TextScaler(text, width, height, mode)])

try:
    # Only create this class when matplotlib is installed
    from matplotlib.transforms import Bbox
    from matplotlib.patches import FancyArrow
    from matplotlib.patheffects import AbstractPathEffect

    class AdaptiveFancyArrow(FancyArrow):
        """
        A `FancyArrow` with fixed head shape.

        The length of the head is proportional to the width of the head
        in display coordinates.
        If the head length is longer than the length of the entire
        arrow, the head length is limited to the arrow length.

        Parameters
        ----------
        x,y  : float
            The arrow's start position.
        dx, dy : float
            The arrow's direction vector, inclduing the arrow head.
        tail_width, head_width : float
            The width of the arrow's tail and head in its reference
            coordinate system
        head_ratio : float, optional
            The length of the arrow head as faction of the arrow width
            (display coordinates).
        draw_head: bool, optional
            If false, the arrow has no head. The result is a rectangle.
        shape : str, optional
            The `shape` parameter in the constructor of `FancyArrow`.
        **kwargs
            Other parameters that are used in the constructor of
            `FancyArrow`.
        """
        
        def __init__(self, x, y, dx, dy,
                     tail_width, head_width, head_ratio, draw_head=True,
                     shape="full", **kwargs):
            import matplotlib.pyplot as plt
            self._x = x
            self._y = y
            self._dx = dx
            self._dy = dy
            self._tail_width = tail_width
            self._head_width = head_width
            self._head_ratio = head_ratio
            self._draw_head = draw_head
            self._shape = shape
            self._kwargs = kwargs
            if not draw_head:
                head_width = tail_width
            super().__init__(
                x, y, dx, dy,
                width=tail_width, head_width=head_width,
                overhang=0, shape=shape,
                length_includes_head=True, **kwargs
            )

        def draw(self, renderer):
            arrow_box = Bbox([(0,0), (0,self._head_width)])
            arrow_box_display = self.axes.transData.transform_bbox(arrow_box)
            head_length_display = np.abs(
                arrow_box_display.height * self._head_ratio
            )
            arrow_box_display.x1 = arrow_box_display.x0 + head_length_display
            # Transfrom back to data coordinates for plotting
            arrow_box = self.axes.transData.inverted().transform_bbox(
                arrow_box_display
            )
            head_length = arrow_box.width
            arrow_length = norm((self._dx, self._dy))
            if head_length > arrow_length:
                # If the head would be longer than the entire arrow,
                # only draw the arrow head with reduced length
                head_length = arrow_length
            if not self._draw_head:
                head_length = 0 

            # Renew the arrow's properties
            super().__init__(
                self._x, self._y, self._dx, self._dy,
                width=self._tail_width, head_width=self._head_width,
                overhang=0, shape=self._shape,
                head_length=head_length, length_includes_head=True,
                axes=self.axes, transform=self.get_transform(), **self._kwargs
            )
            self.set_clip_path(self.axes.patch)
            super().draw(renderer)
        
        # Override to replace docstring
        # Removes warning:
        # unknown document: /tutorials/intermediate/constrainedlayout_guide
        def get_in_layout(self):
            """
            """
            return super().get_in_layout()
        def set_in_layout(self, in_layout):
            """
            """
            return super().set_in_layout(in_layout)
    
    __all__.append("AdaptiveFancyArrow")

except ImportError:
    pass


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.sequence.graphics"
__author__ = "Patrick Kunzmann"
__all__ = ["plot_sequence_logo"]

import numpy as np
from ...visualize import set_font_size_in_coord
from ..alphabet import LetterAlphabet
from .colorschemes import get_color_scheme
import warnings
from ..align import Alignment
from .. import SequenceProfile


def plot_sequence_logo(axes, profile, scheme=None, **kwargs):
    """
    Create a sequence logo. :footcite:`Schneider1990`

    A sequence logo is visualizes the positional composition and
    conservation of a profile encoded in the size of the letters.
    Each position displays all symbols that are occurring at this
    position stacked on each other, with their relative heights depicting
    their relative frequency.
    The height of such a stack depicts its conservation.
    It is the maximum possible Shannon entropy of the alphabet
    subtracted by the positional entropy.

    Parameters
    ----------
    axes : Axes
        The axes to draw the logo one.
    profile: SequenceProfile
        The logo is created based on this profile.
    scheme : str or list of (tuple or str)
        Either a valid color scheme name
        (e.g. ``"rainbow"``, ``"clustalx"``, ``blossom``, etc.)
        or a list of *Matplotlib* compatible colors.
        The list length must be at least as long as the
        length of the alphabet used by the `profile`.
    **kwargs
        Additional `text parameters <https://matplotlib.org/api/text_api.html#matplotlib.text.Text>`_.
    
    References
    ----------
    
    .. footbibliography::
    """
    from matplotlib.text import Text

    if isinstance(profile, Alignment):
        warnings.warn("Using an alignment for this method is deprecated; use a profile instead", DeprecationWarning)
        profile = SequenceProfile.from_alignment(profile)

    alphabet = profile.alphabet
    if not isinstance(alphabet, LetterAlphabet):
        raise TypeError("The sequences' alphabet must be a letter alphabet")

    if scheme is None:
        colors = get_color_scheme("rainbow", alphabet)
    elif isinstance(scheme, str):
        colors = get_color_scheme(scheme, alphabet)
    else:
        colors = scheme
    
    # 'color' and 'size' property is not passed on to text
    kwargs.pop("color", None)
    kwargs.pop("size",  None)
    
    frequencies, entropies, max_entropy = _get_entropy(profile)
    stack_heights = (max_entropy - entropies)
    symbols_heights = stack_heights[:, np.newaxis] * frequencies
    index_order = np.argsort(symbols_heights, axis=1)
    for i in range(symbols_heights.shape[0]):
        # Iterate over the alignment columns
        index_order = np.argsort(symbols_heights)
        start_height = 0
        for j in index_order[i]:
            # Stack the symbols at position on top of the preceeding one
            height = symbols_heights[i,j]
            if height > 0:
                symbol = alphabet.decode(j)
                text = axes.text(
                    i+0.5, start_height, symbol,
                    ha="left", va="bottom", color=colors[j],
                    # Best results are obtained with this font size
                    size=1,
                    **kwargs
                )
                text.set_clip_on(True)
                set_font_size_in_coord(text, width=1, height=height)
                start_height += height

    axes.set_xlim(0.5, len(profile.symbols)+0.5)
    axes.set_ylim(0, max_entropy)


def _get_entropy(profile):
    freq = profile.symbols
    freq = freq / np.sum(freq, axis=1)[:, np.newaxis]
    # 0 * log2(0) = 0 -> Convert NaN to 0
    no_zeros = freq != 0
    pre_entropies = np.zeros(freq.shape)
    pre_entropies[no_zeros] \
        = freq[no_zeros] * np.log2(freq[no_zeros])
    entropies = -np.sum(pre_entropies, axis=1)
    max_entropy = np.log2(len(profile.alphabet))
    return freq, entropies, max_entropy


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite"
__author__ = "Patrick Kunzmann"
__all__ = ["Copyable"]

import abc


class Copyable(metaclass=abc.ABCMeta):
    """
    Base class for all objects, that should be copyable.
    
    The public method `copy()` first creates a fresh instance of the
    class of the instance, that is copied via the `__copy_create__()`
    method. All variables, that could not be set via the constructor,
    are then copied via `__copy_fill__()`, starting with the method in
    the uppermost base class and ending with the class of the instance
    to be copied.
    
    This approach solves the problem of encapsulated variables in
    superclasses.
    """
    
    def copy(self):
        """
        Create a deep copy of this object.
        
        Returns
        -------
        copy
            A copy of this object.
        """
        clone = self.__copy_create__()
        self.__copy_fill__(clone)
        return clone
    
    def __copy_create__(self):
        """
        Instantiate a new object of this class.
        
        Only the constructor should be called in this method.
        All further attributes, that need to be copied are handled
        in `__copy_fill__()`
        
        Do not call the `super()` method here.
        
        This method must be overridden, if the constructor takes
        parameters.
        
        Returns
        -------
        copy
            A freshly instantiated copy of *self*.
        """
        return type(self)()
    
    def __copy_fill__(self, clone):
        """
        Copy all necessary attributes to the new object.
        
        Always call the `super()` method as first statement.
        
        Parameters
        ----------
        clone
            The freshly instantiated copy of *self*.
        """
        pass


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import itertools
import tempfile
import pytest
import biotite.database.uniprot as uniprot
import biotite.sequence.io.fasta as fasta
from biotite.database import RequestError
from ..util import cannot_connect_to


UNIPROT_URL = "https://www.uniprot.org/"


@pytest.mark.skipif(
    cannot_connect_to(UNIPROT_URL),
    reason="UniProt is not available"
)
@pytest.mark.parametrize(
    "as_file_like",
    itertools.product([False, True])
)
def test_fetch(as_file_like):
    path = None if as_file_like else tempfile.gettempdir()

    # UniProtKB
    file = uniprot.fetch(
        "P12345", "fasta", path, overwrite=True
    )
    fasta_file = fasta.FastaFile.read(file)
    prot_seq = fasta.get_sequence(fasta_file)
    assert len(prot_seq) == 430

    # UniRef
    file = uniprot.fetch(
        "UniRef90_P99999", "fasta", path, overwrite=True
    )
    fasta_file = fasta.FastaFile.read(file)
    prot_seq = fasta.get_sequence(fasta_file)
    assert len(prot_seq) == 105

    # UniParc
    file = uniprot.fetch(
        "UPI000000001F", "fasta", path, overwrite=True
    )
    fasta_file = fasta.FastaFile.read(file)
    prot_seq = fasta.get_sequence(fasta_file)
    assert len(prot_seq) == 551


@pytest.mark.skipif(
    cannot_connect_to(UNIPROT_URL),
    reason="UniProt is not available"
)
@pytest.mark.parametrize("format", ["fasta", "gff", "txt", "xml", "rdf", "tab"])
def test_fetch_invalid(format):
    with pytest.raises(RequestError):
        file = uniprot.fetch(
            "xxxx", format, tempfile.gettempdir(), overwrite=True
        )


@pytest.mark.skipif(
    cannot_connect_to(UNIPROT_URL),
    reason="UniProt is not available"
)
def test_search_simple():
    query = uniprot.SimpleQuery("accession", "P62988")
    assert set(uniprot.search(query)) \
        == set(['P62979', 'P0CG47', 'P62987', 'P0CG48', 'P62988'])


@pytest.mark.skipif(
    cannot_connect_to(UNIPROT_URL),
    reason="UniProt is not available"
)
def test_search_composite():
    query = uniprot.SimpleQuery("accession", "P62988") & uniprot.SimpleQuery("reviewed", "yes")
    assert set(uniprot.search(query)) \
        == set(['P62979', 'P0CG47', 'P62987', 'P0CG48'])




# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__author__ = "Daniel Bauer"

import biotite
import pytest


def test_version_number():
    version = biotite.__version__
    assert hasattr(biotite, "__version__")


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
This module provides functions for hydrogen bonding calculation.
"""

__name__ = "biotite.structure"
__author__ = "Daniel Bauer, Patrick Kunzmann"
__all__ = ["hbond", "hbond_frequency"]

import warnings
from .geometry import distance, angle
import numpy as np
from .atoms import AtomArrayStack, stack
from .celllist import CellList


def hbond(atoms, selection1=None, selection2=None, selection1_type='both',
          cutoff_dist=2.5, cutoff_angle=120,
          donor_elements=('O', 'N', 'S'), acceptor_elements=('O', 'N', 'S'),
          periodic=False):
    r"""
    Find hydrogen bonds in a structure using the Baker-Hubbard
    algorithm. :footcite:`Baker1984`

    This function identifies hydrogen bonds based on the bond angle
    :math:`\theta` and the bond distance :math:`d_{H,A}`.
    The default criteria is :math:`\theta > 120^{\circ}`
    and :math:`d_{H,A} \le 2.5 \mathring{A}`.
    Consequently, the given structure must contain hydrogen atoms.
    Otherwise, no hydrogen bonds will be found.
    
    Parameters
    ----------
    atoms : AtomArray or AtomArrayStack
        The atoms to find hydrogen bonds in.
    selection1, selection2: ndarray or None
        Boolean mask for atoms to limit the hydrogen bond search to
        specific sections of the model. The shape must match the
        shape of the `atoms` argument. If None is given, the whole atoms
        stack is used instead. (Default: None)
    selection1_type: {'acceptor', 'donor', 'both'}, optional (default: 'both')
        Determines the type of `selection1`.
        The type of `selection2` is chosen accordingly
        ('both' or the opposite).
        (Default: 'both')
    cutoff_dist : float, optional
        The maximal distance between the hydrogen and acceptor to be
        considered a hydrogen bond. (Default: 2.5)
    cutoff_angle : float, optional
        The angle cutoff in degree between Donor-H..Acceptor to be
        considered a hydrogen bond (default: 120).
    donor_elements, acceptor_elements: tuple of str
        Elements to be considered as possible donors or acceptors
        (Default: O, N, S).
    periodic : bool, optional
        If true, hydrogen bonds can also be detected in periodic
        boundary conditions.
        The `box` attribute of `atoms` is required in this case.
        (Default: False).
        
    Returns
    -------
    triplets : ndarray, dtype=int, shape=(n,3)
        *n x 3* matrix containing the indices of every Donor-H..Acceptor
        interaction that is available in any of the models.
        *n* is the number of found interactions.
        The three matrix columns are *D_index*, *H_index*, *A_index*.
        If only one model (`AtomArray`) is given, `triplets` contains
        all of its hydrogen bonds.
    mask : ndarry, dtype=bool, shape=(m,n)
        *m x n* matrix that shows if an interaction with index *n* in
        `triplets` is present in the model *m* of the input `atoms`.
        Only returned if `atoms` is an :class:`AtomArrayStack`.
    
    Notes
    -----
    The result of this function may include false positives:
    Only the chemical elements and the bond geometry is checked.
    However, there are some cases where a hydrogen bond is still not
    reasonable.
    For example, a nitrogen atom with positive charge could be
    considered as acceptor atom by this method, although this does
    make sense from a chemical perspective.
        
    Examples
    --------
    Calculate the total number of hydrogen bonds found in each model:
    
    >>> triplets, mask = hbond(atom_array_stack)
    >>> hbonds_per_model = np.count_nonzero(mask, axis=1)
    >>> print(hbonds_per_model)
    [14 15 15 13 11 13  9 14  9 15 13 13 15 11 11 13 11 14 14 13 14 13 15 17
     14 12 15 12 12 13 13 13 12 12 11 15 10 11]
    
    Get hydrogen bond donors of third model:
    
    >>> # Third model -> index 2
    >>> triplets = triplets[mask[2,:]]
    >>> # First column contains donors
    >>> print(atom_array_stack[2, triplets[:,0]])
        A       1  ASN N      N        -6.589    7.754   -0.571
        A       5  GLN N      N        -5.009   -0.575   -1.365
        A       6  TRP N      N        -2.154   -0.497   -1.588
        A       7  LEU N      N        -1.520   -1.904    0.893
        A       8  LYS N      N        -2.716   -4.413    0.176
        A       8  LYS NZ     N        -6.352   -4.311   -4.482
        A       9  ASP N      N        -0.694   -5.301   -1.644
        A      11  GLY N      N         2.142   -4.244    1.916
        A      10  GLY N      N         1.135   -6.232    0.250
        A      14  SER OG     O         4.689   -5.759   -2.390
        A      13  SER N      N         6.424   -5.220    3.257
        A      14  SER N      N         6.424   -5.506    0.464
        A      15  GLY N      N         8.320   -3.632   -0.318
        A      16  ARG N      N         8.043   -1.206   -1.866
        A       6  TRP NE1    N         3.420    0.332   -0.121

    See Also
    --------
    hbond_frequency

    References
    ----------

    .. footbibliography::
    """
    if not (atoms.element == "H").any():
        warnings.warn(
            "Input structure does not contain hydrogen atoms, "
            "hence no hydrogen bonds can be identified"
        )

    # Create AtomArrayStack from AtomArray
    if not isinstance(atoms, AtomArrayStack):
        atoms = stack([atoms])
        single_model = True
    else:
        single_model = False
    
    if periodic:
        box = atoms.box
    else:
        box = None
    
    # Mask for donor/acceptor elements
    donor_element_mask = np.isin(atoms.element, donor_elements)
    acceptor_element_mask = np.isin(atoms.element, acceptor_elements)

    if selection1 is None:
        selection1 = np.ones(atoms.array_length(), dtype=bool)
    if selection2 is None:
        selection2 = np.ones(atoms.array_length(), dtype=bool)

    if selection1_type == 'both':
        # The two selections are separated into three selections:
        # the original ones without the overlaping part
        # and one containing the overlap
        # This prevents redundant triplets and unnecessary computation 
        overlap_selection = selection1 & selection2
        # Original selections without overlaping part
        exclusive_selection1 = selection1 & (~overlap_selection)
        exclusive_selection2 = selection2 & (~overlap_selection)
        
        # Put selections to list for cleaner iteration
        selections = [
            exclusive_selection1, exclusive_selection2, overlap_selection
        ]
        selection_combinations = [
            #(0,0),   is not included, would be same selection
            #         as donor and acceptor simultaneously
            (0,1),
            (0,2),
            (1,0),
            #(1,1),   # same reason above
            (1,2),
            (2,0),
            (2,1),
            (2,2)     # overlaping part, combination is necessary
        ]
        
        all_comb_triplets = []
        all_comb_mask = []
        for selection_index1, selection_index2 in selection_combinations:
            donor_mask = selections[selection_index1]
            acceptor_mask = selections[selection_index2]
            if  np.count_nonzero(donor_mask) != 0 and \
                np.count_nonzero(acceptor_mask) != 0:
                    # Calculate triplets and mask
                    triplets, mask = _hbond(
                        atoms, donor_mask, acceptor_mask,
                        donor_element_mask, acceptor_element_mask,
                        cutoff_dist, cutoff_angle,
                        donor_elements, acceptor_elements,
                        box
                    )
                    all_comb_triplets.append(triplets)
                    all_comb_mask.append(mask)
        # Merge results from all combinations
        triplets = np.concatenate(all_comb_triplets, axis=0)
        mask = np.concatenate(all_comb_mask, axis=1)

    elif selection1_type == 'donor':
        triplets, mask = _hbond(
            atoms, selection1, selection2,
            donor_element_mask, acceptor_element_mask,
            cutoff_dist, cutoff_angle,
            donor_elements, acceptor_elements,
            box
        )
    
    elif selection1_type == 'acceptor':
        triplets, mask = _hbond(
            atoms, selection2, selection1,
            donor_element_mask, acceptor_element_mask,
            cutoff_dist, cutoff_angle,
            donor_elements, acceptor_elements,
            box
        )
    
    else:
        raise ValueError(f"Unkown selection type '{selection1_type}'")

    if single_model:
        # For a atom array (not stack),
        # hbond_mask contains only 'True' values,
        # since all interaction are in the one model
        # -> Simply return triplets without hbond_mask
        return triplets
    else:
        return triplets, mask


def _hbond(atoms, donor_mask, acceptor_mask,
           donor_element_mask, acceptor_element_mask,
           cutoff_dist, cutoff_angle, donor_elements, acceptor_elements,
           box):
    
    # Filter donor/acceptor elements
    donor_mask    &= donor_element_mask
    acceptor_mask &= acceptor_element_mask
    
    def _get_bonded_hydrogens(array, donor_mask, box, cutoff=1.5):
        """
        Helper function to find indices of associated hydrogens in atoms
        for all donors in atoms[donor_mask].
        The criterium is that the hydrogen must be in the same residue
        and the distance must be smaller than the cutoff.

        """
        coord = array.coord
        res_id = array.res_id
        hydrogen_mask = (array.element == "H")
        
        donor_hydrogen_mask = np.zeros(len(array), dtype=bool)
        associated_donor_indices = np.full(len(array), -1, dtype=int)

        donor_indices = np.where(donor_mask)[0]
        for donor_i in donor_indices:
            candidate_mask = hydrogen_mask & (res_id == res_id[donor_i])
            distances = distance(
                coord[donor_i], coord[candidate_mask], box=box
            )
            donor_h_indices = np.where(candidate_mask)[0][distances <= cutoff]
            for i in donor_h_indices:
                associated_donor_indices[i] = donor_i
                donor_hydrogen_mask[i] = True
        
        return donor_hydrogen_mask, associated_donor_indices
            

    # TODO use BondList if available
    first_model_box = box[0] if box is not None else None
    donor_h_mask, associated_donor_indices \
        = _get_bonded_hydrogens(atoms[0], donor_mask, first_model_box)
    donor_h_i = np.where(donor_h_mask)[0]
    acceptor_i = np.where(acceptor_mask)[0]
    if len(donor_h_i) == 0 or len(acceptor_i) == 0:
        # Return empty triplets and mask
        return (
            np.zeros((0,3), dtype=int),
            np.zeros((atoms.stack_depth(),0), dtype=bool)
        )
    
    # Narrow the amount of possible acceptor to donor-H connections
    # down via the distance cutoff parameter using a cell list
    # Save in acceptor-to-hydrogen matrix
    # (true when distance smaller than cutoff)
    coord = atoms.coord
    possible_bonds = np.zeros(
        (len(acceptor_i), len(donor_h_i)),
        dtype=bool
    )
    periodic = False if box is None else True
    for model_i in range(atoms.stack_depth()):
        donor_h_coord = coord[model_i, donor_h_mask]
        acceptor_coord = coord[model_i, acceptor_mask]
        box_for_model = box[model_i] if box is not None else None
        cell_list = CellList(
            donor_h_coord, cell_size=cutoff_dist,
            periodic=periodic, box=box_for_model
        )
        possible_bonds |= cell_list.get_atoms_in_cells(
            acceptor_coord, as_mask=True
        )
    possible_bonds_i = np.where(possible_bonds)
    # Narrow down
    acceptor_i = acceptor_i[possible_bonds_i[0]]
    donor_h_i = donor_h_i[possible_bonds_i[1]]
    
    # Build D-H..A triplets
    donor_i = associated_donor_indices[donor_h_i]
    triplets = np.stack((donor_i, donor_h_i, acceptor_i), axis=1)
    # Remove entries where donor and acceptor are the same
    triplets = triplets[donor_i != acceptor_i]

     # Filter triplets that meet distance and angle condition
    def _is_hbond(donor, donor_h, acceptor, box,
                  cutoff_dist=2.5, cutoff_angle=120):
        cutoff_angle_rad = np.deg2rad(cutoff_angle)
        theta = angle(donor, donor_h, acceptor, box=box)
        dist = distance(donor_h, acceptor, box=box)
        return (theta > cutoff_angle_rad) & (dist <= cutoff_dist)
    
    hbond_mask = _is_hbond(
        coord[:, triplets[:,0]],  # donors
        coord[:, triplets[:,1]],  # donor hydrogens
        coord[:, triplets[:,2]],  # acceptors
        box, cutoff_dist=cutoff_dist, cutoff_angle=cutoff_angle
    )

    # Reduce output to contain only triplets counted at least once
    is_counted = hbond_mask.any(axis=0)
    triplets = triplets[is_counted]
    hbond_mask = hbond_mask[:, is_counted]

    return triplets, hbond_mask


def hbond_frequency(mask):
    """
    Get the relative frequency of each hydrogen bond in a multi-model
    structure.

    The frequency is the amount of models, where the respective bond
    exists divided by the total amount of models.
    
    Parameters
    ----------
    mask: ndarray, dtype=bool, shape=(m,n)
        Input mask obtained from `hbond` function.
    
    Returns
    -------
    ndarray, dtype=Float
        For each individual interaction *n* of the mask, returns the
        percentage of models *m*, in which this hydrogen bond is
        present.

    See Also
    --------
    hbond

    Examples
    --------

    >>> triplets, mask = hbond(atom_array_stack)
    >>> freq = hbond_frequency(mask)
    >>> print(freq)
    [0.263 0.289 0.105 0.105 0.237 0.237 0.026 0.053 0.395 1.000 1.000 1.000
     0.026 0.421 0.026 0.026 0.316 0.816 0.026 0.921 0.026 0.342 0.026 0.105
     0.026 0.132 0.053 0.026 0.158 0.026 0.868 0.211 0.026 0.921 0.316 0.079
     0.237 0.105 0.421 0.079 0.026 1.000 0.053 0.132 0.026 0.184]
    """
    return mask.sum(axis=0)/len(mask)



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import biotite.structure as struc
import biotite.structure.io as strucio
import numpy as np
from os.path import join
from ..util import data_dir
import pytest

@pytest.fixture
def sample_protein():
    return strucio.load_structure(
        join(data_dir("structure"), "3o5r.mmtf")
    )

@pytest.fixture
def sample_nucleotide():
    return strucio.load_structure(
        join(data_dir("structure"), "5ugo.mmtf")
    )

@pytest.fixture
def all_atloc_structure():
    return strucio.load_structure(
        join(data_dir("structure"), "1o1z.mmtf"),
        extra_fields = ["occupancy"],
        altloc="all"
    )

def test_solvent_filter(sample_protein):
    assert len(sample_protein[struc.filter_solvent(sample_protein)]) == 287

def test_amino_acid_filter(sample_protein):
    assert len(sample_protein[struc.filter_amino_acids(sample_protein)]) == 982

def test_backbone_filter(sample_protein):
    assert len(sample_protein[struc.filter_backbone(sample_protein)]) == 384

def test_intersection_filter(sample_protein):
    assert len(sample_protein[:200][
        struc.filter_intersection(sample_protein[:200],sample_protein[100:])
    ]) == 100

def test_nucleotide_filter(sample_nucleotide):

    assert len(
        sample_nucleotide[struc.filter_nucleotides(sample_nucleotide)]
    ) == 651


@pytest.mark.parametrize("filter_func", ["first", "occupancy"])
def test_filter_altloc(all_atloc_structure, filter_func):
    """
    For a correctly altloc filtered structure no atom should be missing
    and no atom should be present twice.
    """
    ref_atom_set = set()
    for atom_tuple in zip(
        all_atloc_structure.chain_id,
        all_atloc_structure.res_id,
        all_atloc_structure.ins_code,
        all_atloc_structure.atom_name
    ):
        ref_atom_set.add(atom_tuple)
    
    if filter_func == "first":
        filtered_structure = all_atloc_structure[struc.filter_first_altloc(
            all_atloc_structure,
            all_atloc_structure.altloc_id
        )]
    elif filter_func == "occupancy":
        filtered_structure = all_atloc_structure[
            struc.filter_highest_occupancy_altloc(
                all_atloc_structure,
                all_atloc_structure.altloc_id,
                all_atloc_structure.occupancy
            )
        ]

    test_atom_set = set()
    for atom_tuple in zip(
        filtered_structure.chain_id,
        filtered_structure.res_id,
        filtered_structure.ins_code,
        filtered_structure.atom_name
    ):
        try:
            # No atom should be present twice
            assert atom_tuple not in test_atom_set
        except AssertionError:
            print(f"Atom {atom_tuple} is present twice")
            raise
        test_atom_set.add(atom_tuple)
    
    # No atom should be missing
    assert test_atom_set == ref_atom_set


def test_filter_highest_occupancy_altloc(all_atloc_structure):
    """
    When filtering altlocs with the highest occupancy, the average
    occupancy should be higher than the the average occupancy when the
    first altloc ID is always filtered.
    """
    # Set the occupancy of SECOND altloc ID very high
    all_atloc_structure.occupancy[all_atloc_structure.altloc_id == "B"] = 1.0
    
    # filter_first_altloc
    filtered_structure = all_atloc_structure[struc.filter_first_altloc(
        all_atloc_structure,
        all_atloc_structure.altloc_id
    )]
    ref_occupancy_sum = np.average(filtered_structure.occupancy)

    # filter_highest_occupancy_altloc
    filtered_structure = all_atloc_structure[
        struc.filter_highest_occupancy_altloc(
            all_atloc_structure,
            all_atloc_structure.altloc_id,
            all_atloc_structure.occupancy
        )
    ]
    test_occupancy_sum = np.average(filtered_structure.occupancy)
    
    assert test_occupancy_sum > ref_occupancy_sum


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
This module handles conversion of RNA structures to
dot-bracket-notation.
"""

__name__ = "biotite.structure"
__author__ = "Tom David Müller"
__all__ = ["dot_bracket_from_structure", "dot_bracket",
           "base_pairs_from_dot_bracket"]

import numpy as np
from .basepairs import base_pairs
from .pseudoknots import pseudoknots
from .residues import get_residue_count, get_residue_positions

_OPENING_BRACKETS = "([{<ABCDEFGHIJKLMNOPQRSTUVWXYZ"
_OPENING_BRACKETS_BYTES = _OPENING_BRACKETS.encode()
_CLOSING_BRACKETS = ")]}>abcdefghijklmnopqrstuvwxyz"
_CLOSING_BRACKETS_BYTES = _CLOSING_BRACKETS.encode()


def dot_bracket_from_structure(
    nucleic_acid_strand, scores=None, max_pseudoknot_order=None):
    """
    Represent a nucleic-acid-strand in dot-bracket-letter-notation
    (DBL-notation). :footcite:`Antczak2018`

    Parameters
    ----------
    atom_array : AtomArray
        The nucleic acid strand to be represented in DBL-notation.
    scores : ndarray, dtype=int, shape=(n,) (default: None)
        The score for each base pair, which is passed on to
        :func:`pseudoknots()`.
    max_pseudoknot_order : int (default: None)
        The maximum pseudoknot order to be found. If a base pair would
        be of a higher order, it is represented as unpaired. If ``None``
        is given, all base pairs are evaluated.

    Returns
    -------
    notations : list [str, ...]
        The DBL-notation for each solution from :func:`pseudoknots()`.

    See Also
    --------
    base_pairs
    pseudoknots

    References
    ----------
    
    .. footbibliography::
    """
    basepairs = base_pairs(nucleic_acid_strand)
    basepairs = get_residue_positions(nucleic_acid_strand, basepairs)
    length = get_residue_count(nucleic_acid_strand)
    return dot_bracket(basepairs, length, scores=scores,
                       max_pseudoknot_order=max_pseudoknot_order)

def dot_bracket(basepairs, length, scores=None, max_pseudoknot_order=None):
    """
    Represent a nucleic acid strand in dot-bracket-letter-notation
    (DBL-notation). :footcite:`Antczak2018`

    The nucleic acid strand is represented as nucleotide sequence,
    where the nucleotides are counted continiously from zero.

    Parameters
    ----------
    basepairs : ndarray, shape=(n,2)
        Each row corresponds to the positions of the bases in the
        strand.
    length : int
        The number of bases in the strand.
    scores : ndarray, dtype=int, shape=(n,) (default: None)
        The score for each base pair, which is passed on to
        :func:`pseudoknots()`
    max_pseudoknot_order : int (default: None)
        The maximum pseudoknot order to be found. If a base pair would
        be of a higher order, it is represented as unpaired. If ``None``
        is given, all pseudoknot orders are evaluated.

    Returns
    -------
    notations : list [str, ...]
        The DBL-notation for each solution from :func:`pseudoknots()`.

    Examples
    --------
    The sequence ``ACGTC`` has a length of 5. If there was to be a
    pairing interaction between the ``A`` and ``T``, `basepairs` would
    have the form:

    >>> import numpy as np
    >>> basepairs = np.array([[0, 3]])

    The DBL Notation can then be found with ``dot_bracket()``:

    >>> dot_bracket(basepairs, 5)[0]
    '(..).'


    See Also
    --------
    dot_bracket_from_structure
    base_pairs
    pseudoknots

    References
    ----------
    
    .. footbibliography::
    """
    # Make sure the lower residue is on the left for each row
    basepairs = np.sort(basepairs, axis=1)

    # Get pseudoknot order
    pseudoknot_order = pseudoknots(basepairs, scores=scores,
                                   max_pseudoknot_order=max_pseudoknot_order)

    # Each optimal pseudoknot order solution is represented in
    # dot-bracket-notation
    notations = [
        bytearray((b"."*length)) for _ in range(len(pseudoknot_order))
    ]
    for s, solution in enumerate(pseudoknot_order):
        for basepair, order in zip(basepairs, solution):
            if order == -1:
                continue
            notations[s][basepair[0]] = _OPENING_BRACKETS_BYTES[order]
            notations[s][basepair[1]] = _CLOSING_BRACKETS_BYTES[order]
    return [notation.decode() for notation in notations]

def base_pairs_from_dot_bracket(dot_bracket_notation):
    """
    Extract the base pairs from a nucleic-acid-strand in
    dot-bracket-letter-notation (DBL-notation). :footcite:`Antczak2018`

    The nucleic acid strand is represented as nucleotide sequence,
    where the nucleotides are counted continiously from zero.

    Parameters
    ----------
    dot_bracket_notation : str
        The DBL-notation.

    Returns
    -------
    basepairs : ndarray, shape=(n,2)
        Each row corresponds to the positions of the bases in the
        sequence.

    Examples
    --------
    The notation string ``'(..).'`` contains a base pair between the
    indices 0 and 3. This pairing interaction can be extracted
    conveniently by the use of :func:`base_pairs_from_dot_bracket()`:

    >>> base_pairs_from_dot_bracket('(..).')
    array([[0, 3]])

    See Also
    --------
    dot_bracket

    References
    ----------
    
    .. footbibliography::
    """
    basepairs = []
    opened_brackets = [[] for _ in range(len(_OPENING_BRACKETS))]

    # Iterate through input string and extract base pairs
    for pos, symbol in enumerate(dot_bracket_notation):

        if symbol in _OPENING_BRACKETS:
            # Add opening residues to list (separate list for each
            # bracket type)
            index = _OPENING_BRACKETS.index(symbol)
            opened_brackets[index].append(pos)

        elif symbol in _CLOSING_BRACKETS:
            # For each closing bracket, the the base pair consists out
            # of the current index and the last index added to the list
            # in `opened_brackets` corresponding to the same bracket
            # type.
            index = _CLOSING_BRACKETS.index(symbol)
            basepairs.append((opened_brackets[index].pop(), pos))

        else:
            if symbol != ".":
                raise ValueError(
                    f"'{symbol}' is an invalid character for DBL-notation"
                )

    for not_closed in opened_brackets:
        if not_closed != []:
            raise ValueError(
                "Invalid DBL-notation, not all opening brackets have a "
                "closing bracket"
            )


    # Sort the base pair indices in ascending order
    basepairs = np.array(basepairs)
    basepairs = basepairs[np.argsort(basepairs[:, 0])]
    return basepairs



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
This module provides functions for geometric measurements between atoms
in a structure, mainly lenghts and angles.
"""

__name__ = "biotite.structure"
__author__ = "Patrick Kunzmann"
__all__ = ["displacement", "index_displacement", "distance", "index_distance",
           "angle", "index_angle", "dihedral", "index_dihedral",
           "dihedral_backbone", "centroid"]

import numpy as np
from .atoms import Atom, AtomArray, AtomArrayStack, coord
from .util import vector_dot, norm_vector
from .filter import filter_backbone
from .chains import chain_iter
from .box import (coord_to_fraction, fraction_to_coord,
                  move_inside_box, is_orthogonal)
from .error import BadStructureError


def displacement(atoms1, atoms2, box=None):
    """
    Measure the displacement vector, i.e. the vector difference, from
    one array of atom coordinates to another array of coordinates.
    
    Parameters
    ----------
    atoms1, atoms2 : ndarray, shape=(m,n,3) or ndarray, shape=(n,3) or ndarray, shape=(3,) or Atom or AtomArray or AtomArrayStack
        The atoms to measure the displacement between.
        The vector from `atoms1` to `atoms2` is measured.
        The dimensions may vary.
        Alternatively an ndarray containing the coordinates can be
        provided.
        Usual *NumPy* broadcasting rules apply.
    box : ndarray, shape=(3,3) or shape=(m,3,3), optional
        If this parameter is set, periodic boundary conditions are
        taken into account (minimum-image convention), based on
        the box vectors given with this parameter.
        The shape *(m,3,3)* is only allowed, when the input coordinates
        comprise multiple models.
    
    Returns
    -------
    disp : ndarray, shape=(m,n,3) or ndarray, shape=(n,3) or ndarray, shape=(3,)
        The displacement vector(s). The shape is equal to the shape of
        the input `atoms` with the highest dimensionality.
    
    See also
    --------
    index_displacement
    """
    v1 = coord(atoms1)
    v2 = coord(atoms2)
    # Decide subtraction order based on shape, since an array can be
    # only subtracted by an array with less dimensions
    if len(v1.shape) <= len(v2.shape):
        diff = v2 - v1
    else:
        diff = -(v1 - v2)
    
    # Use minimum-image convention if box is given
    if box is not None:
        # Transform difference vector
        # from coordinates into fractions of box vectors
        # for faster calculation laster on
        fractions = coord_to_fraction(diff, box)
        # Move vectors into box
        fractions = fractions % 1
        # Check for each model if the box vectors are orthogonal
        orthogonality = is_orthogonal(box)
        disp = np.zeros(fractions.shape, dtype=diff.dtype)
        if fractions.ndim == 1:
            # Single atom
            # Transform into two dimensions
            # to match signature of '_displacement_xxx()'
            fractions = fractions[np.newaxis, :]
            disp = disp[np.newaxis, :]
            if orthogonality:
                _displacement_orthogonal_box(
                    fractions, box, disp
                )
            else:
                _displacement_triclinic_box(
                    fractions.astype(diff.dtype, copy=False),
                    box.astype(diff.dtype, copy=False),
                    disp
                )
            # Transform back
            disp = disp[0]
        if fractions.ndim == 2:
            # Single model
            if orthogonality:
                _displacement_orthogonal_box(
                    fractions, box, disp
                )
            else:
                _displacement_triclinic_box(
                    fractions.astype(diff.dtype, copy=False),
                    box.astype(diff.dtype, copy=False),
                    disp
                )
        elif fractions.ndim == 3:
            # Multiple models
            # (Model count) x (Atom count)
            for i in range(len(fractions)):
                if box.ndim == 2:
                    box_for_model = box
                    orthogonality_for_model = orthogonality
                elif box.ndim == 3:
                    box_for_model = box[i]
                    orthogonality_for_model = orthogonality[i]
                else:
                    raise ValueError(f"{box.ndim} are to many box dimensions")
                if orthogonality_for_model:
                    _displacement_orthogonal_box(
                        fractions[i], box_for_model, disp[i]
                    )
                else:
                    _displacement_triclinic_box(
                        fractions[i].astype(diff.dtype, copy=False),
                        box_for_model.astype(diff.dtype, copy=False),
                        disp[i]
                    )
        else:
            raise ValueError(
                f"{diff.shape} is an invalid shape for atom coordinates"
            )
        return disp
    
    else:
        return diff


def index_displacement(*args, **kwargs):
    """
    index_displacement(atoms, indices, periodic=False, box=None)
    
    Measure the displacement, i.e. the vector difference, between pairs
    of atoms.

    The pairs refer to indices of a given atom array, whose pairwise
    displacement should be calculated.
    If an atom array stack is provided, the distances are calculated for
    each frame/model.
    In contrast to the :func:`distance()` function, this function is
    able to take periodic boundary conditions into account.

    Parameters
    ----------
    atoms : AtomArray or AtomArrayStack or ndarray, shape=(n,3) or shape=(m,n,3)
        The atoms the `indices` parameter refers to.
        The pairwise distances are calculated for these pairs.
        Alternatively, the atom coordinates can be directly provided as
        :class:`ndarray`.
    indices : ndarray, shape=(k,2)
        Pairs of indices that point to `atoms`.
        The displacement is measured from ``indices[x,0]`` to 
        ``indices[x,1]``.
    periodic : bool, optional
        If set to true, periodic boundary conditions are taken into
        account (minimum-image convention).
        The `box` attribute of the `atoms` parameter is used for
        calculation.
        An alternative box can be provided via the `box` parameter.
        By default, periodicity is ignored.
    box : ndarray, shape=(3,3) or shape=(m,3,3), optional
        If this parameter is set, the given box is used instead of the
        `box` attribute of `atoms`.
    
    Returns
    -------
    disp : ndarray, shape=(k,) or shape=(m,k)
        The pairwise displacements.
        If `atoms` is an atom array stack, The distances are
        calculated for each model.
    
    Warnings
    --------
    In case `periodic` is set to true and if the box is not orthorhombic
    (at least one angle deviates from 90 degrees),
    the calculation requires approximately 8 times as long as in the
    orthorhombic case.
    Furthermore, it is not guaranteed, that the lowest-distance periodic
    copy is found for non-orthorhombic boxes; this is especially true
    for heavily skewed boxes.

    See also
    --------
    displacement
    """
    return _call_non_index_function(displacement, 2, *args, **kwargs)


def distance(atoms1, atoms2, box=None):
    """
    Measure the euclidian distance between atoms.
    
    Parameters
    ----------
    atoms1, atoms2 : ndarray or Atom or AtomArray or AtomArrayStack
        The atoms to measure the distances between.
        The dimensions may vary.
        Alternatively an ndarray containing the coordinates can be
        provided.
        Usual *NumPy* broadcasting rules apply.
    box : ndarray, shape=(3,3) or shape=(m,3,3), optional
        If this parameter is set, periodic boundary conditions are
        taken into account (minimum-image convention), based on
        the box vectors given with this parameter.
        The shape *(m,3,3)* is only allowed, when the input coordinates
        comprise multiple models.
    
    Returns
    -------
    dist : float or ndarray
        The atom distances.
        The shape is equal to the shape of the input `atoms` with the
        highest dimensionality minus the last axis.
    
    See also
    --------
    index_distance
    """
    diff = displacement(atoms1, atoms2, box)
    return np.sqrt(vector_dot(diff, diff))


def index_distance(*args, **kwargs):
    """
    index_distance(atoms, indices, periodic=False, box=None)
    
    Measure the euclidian distance between pairs of atoms.

    The pairs refer to indices of a given atom array, whose pairwise
    distances should be calculated.
    If an atom array stack is provided, the distances are calculated for
    each frame/model.
    In contrast to the :func:`distance()` function, this function is
    able to take periodic boundary conditions into account.

    Parameters
    ----------
    atoms : AtomArray or AtomArrayStack or ndarray, shape=(n,3) or shape=(m,n,3)
        The atoms the `indices` parameter refers to.
        The pairwise distances are calculated for these pairs.
        Alternatively, the atom coordinates can be directly provided as
        :class:`ndarray`.
    indices : ndarray, shape=(k,2)
        Pairs of indices that point to `atoms`.
    periodic : bool, optional
        If set to true, periodic boundary conditions are taken into
        account (minimum-image convention).
        The `box` attribute of the `atoms` parameter is used for
        calculation.
        An alternative box can be provided via the `box` parameter.
        By default, periodicity is ignored.
    box : ndarray, shape=(3,3) or shape=(m,3,3), optional
        If this parameter is set, the given box is used instead of the
        `box` attribute of `atoms`.
    
    Returns
    -------
    dist : ndarray, shape=(k,) or shape=(m,k)
        The pairwise distances.
        If `atoms` is an atom array stack, The distances are
        calculated for each model.
    
    Warnings
    --------
    In case `periodic` is set to true and if the box is not orthorhombic
    (at least one angle deviates from 90 degrees),
    the calculation requires approximately 8 times as long as in the
    orthorhombic case.
    Furthermore, it is not guaranteed, that the lowest-distance periodic
    copy is found for non-orthorhombic boxes; this is especially true
    for heavily skewed boxes.

    See also
    --------
    distance
    """
    return _call_non_index_function(distance, 2, *args, **kwargs)


def angle(atoms1, atoms2, atoms3, box=None):
    """
    Measure the angle between 3 atoms.
    
    Parameters
    ----------
    atoms1, atoms2, atoms3 : ndarray or Atom or AtomArray or AtomArrayStack
        The atoms to measure the angle between. Alternatively an
        ndarray containing the coordinates can be provided.
    
    Returns
    -------
    angle : float or ndarray
        The angle(s) between the atoms. The shape is equal to the shape
        of the input `atoms` with the highest dimensionality minus the
        last axis.
    box : ndarray, shape=(3,3) or shape=(m,3,3), optional
        If this parameter is set, periodic boundary conditions are
        taken into account (minimum-image convention), based on
        the box vectors given with this parameter.
        The shape *(m,3,3)* is only allowed, when the input coordinates
        comprise multiple models.
    
    See also
    --------
    index_angle
    """
    v1 = displacement(atoms1, atoms2, box)
    v2 = displacement(atoms3, atoms2, box)
    norm_vector(v1)
    norm_vector(v2)
    return np.arccos(vector_dot(v1,v2))


def index_angle(*args, **kwargs):
    """
    index_angle(atoms, indices, periodic=False, box=None)
    
    Measure the angle between triples of atoms.

    The triples refer to indices of a given atom array, whose triplewise
    angles should be calculated.
    If an atom array stack is provided, the distances are calculated for
    each frame/model.

    Parameters
    ----------
    atoms : AtomArray or AtomArrayStack or ndarray, shape=(n,3) or shape=(m,n,3)
        The atoms the `indices` parameter refers to.
        The triplewise distances are calculated for these pairs.
        Alternatively, the atom coordinates can be directly provided as
        :class:`ndarray`.
    indices : ndarray, shape=(k,3)
        Triples of indices that point to `atoms`.
    periodic : bool, optional
        If set to true, periodic boundary conditions are taken into
        account (minimum-image convention).
        The `box` attribute of the `atoms` parameter is used for
        calculation.
        An alternative box can be provided via the `box` parameter.
        By default, periodicity is ignored.
    box : ndarray, shape=(3,3) or shape=(m,3,3), optional
        If this parameter is set, the given box is used instead of the
        `box` attribute of `atoms`.
    
    Returns
    -------
    angle : ndarray, shape=(k,) or shape=(m,k)
        The triplewise angles.
        If `atoms` is an atom array stack, The distances are
        calculated for each model.
    
    Warnings
    --------
    In case `periodic` is set to true and if the box is not orthorhombic
    (at least one angle deviates from 90 degrees),
    the calculation requires approximately 8 times as long as in the
    orthorhombic case.
    Furthermore, it is not guaranteed, that the lowest-distance periodic
    copy is found for non-orthorhombic boxes; this is especially true
    for heavily skewed boxes.

    See also
    --------
    angle
    """
    return _call_non_index_function(angle, 3, *args, **kwargs)


def dihedral(atoms1, atoms2, atoms3, atoms4, box=None):
    """
    Measure the dihedral angle between 4 atoms.
    
    Parameters
    ----------
    atoms1, atoms2, atoms3, atoms4 : ndarray or Atom or AtomArray or AtomArrayStack
        The atoms to measure the dihedral angle between.
        Alternatively an ndarray containing the coordinates can be
        provided.
    
    Returns
    -------
    dihed : float or ndarray
        The dihedral angle(s) between the atoms. The shape is equal to
        the shape of the input `atoms` with the highest dimensionality
        minus the last axis.
    box : ndarray, shape=(3,3) or shape=(m,3,3), optional
        If this parameter is set, periodic boundary conditions are
        taken into account (minimum-image convention), based on
        the box vectors given with this parameter.
        The shape *(m,3,3)* is only allowed, when the input coordinates
        comprise multiple models.
    
    See Also
    --------
    index_dihedral
    dihedral_backbone
    """
    v1 = displacement(atoms1, atoms2, box)
    v2 = displacement(atoms2, atoms3, box)
    v3 = displacement(atoms3, atoms4, box)
    norm_vector(v1)
    norm_vector(v2)
    norm_vector(v3)
    
    n1 = np.cross(v1, v2)
    n2 = np.cross(v2, v3)
    
    # Calculation using atan2, to ensure the correct sign of the angle 
    x = vector_dot(n1,n2)
    y = vector_dot(np.cross(n1,n2), v2)
    return np.arctan2(y,x)


def index_dihedral(*args, **kwargs):
    """
    index_dihedral(atoms, indices, periodic=False, box=None)
    
    Measure the dihedral angle between quadruples of atoms.

    The triples refer to indices of a given atom array, whose
    quadruplewise dihedral angles should be calculated.
    If an atom array stack is provided, the distances are calculated for
    each frame/model.

    Parameters
    ----------
    atoms : AtomArray or AtomArrayStack or ndarray, shape=(n,3) or shape=(m,n,3)
        The atoms the `indices` parameter refers to.
        The quadruplewise dihedral angles are calculated for these
        pairs.
        Alternatively, the atom coordinates can be directly provided as
        :class:`ndarray`.
    indices : ndarray, shape=(k,4)
        Quadruples of indices that point to `atoms`.
    periodic : bool, optional
        If set to true, periodic boundary conditions are taken into
        account (minimum-image convention).
        The `box` attribute of the `atoms` parameter is used for
        calculation.
        An alternative box can be provided via the `box` parameter.
        By default, periodicity is ignored.
    box : ndarray, shape=(3,3) or shape=(m,3,3), optional
        If this parameter is set, the given box is used instead of the
        `box` attribute of `atoms`.
    
    Returns
    -------
    dihedral : ndarray, shape=(k,) or shape=(m,k)
        The quadruplewise dihedral angles.
        If `atoms` is an atom array stack, The distances are
        calculated for each model.
    
    Warnings
    --------
    In case `periodic` is set to true and if the box is not orthorhombic
    (at least one angle deviates from 90 degrees),
    the calculation requires approximately 8 times as long as in the
    orthorhombic case.
    Furthermore, it is not guaranteed, that the lowest-distance periodic
    copy is found for non-orthorhombic boxes; this is especially true
    for heavily skewed boxes.

    See also
    --------
    dihedral
    dihedral_backbone
    """
    return _call_non_index_function(dihedral, 4, *args, **kwargs)


def dihedral_backbone(atom_array):
    """
    Measure the characteristic backbone dihedral angles of a protein
    structure.
    
    Parameters
    ----------
    atom_array: AtomArray or AtomArrayStack
        The protein structure. A complete backbone, without gaps,
        is required here.
        Chain transitions are allowed, the angles at the transition are
        `NaN`.
        The order of the backbone atoms for each residue must be
        (N, CA, C).
    
    Returns
    -------
    phi, psi, omega : ndarray
        An array containing the 3 backbone dihedral angles for every
        CA. 'phi' is not defined at the N-terminus, 'psi' and 'omega'
        are not defined at the C-terminus. In these places the arrays
        have *NaN* values. If an :class:`AtomArrayStack` is given, the
        output angles are 2-dimensional, the first dimension corresponds
        to the model number.
    
    Raises
    ------
    BadStructureError
        If the amount of backbone atoms is not equal to amount of
        residues times 3 (for N, CA and C).
    
    See Also
    --------
    dihedral
    
    Examples
    --------
    
    >>> phi, psi, omega = dihedral_backbone(atom_array)
    >>> print(np.stack([np.rad2deg(phi), np.rad2deg(psi)]).T)
    [[     nan  -56.145]
     [ -43.980  -51.309]
     [ -66.466  -30.898]
     [ -65.219  -45.945]
     [ -64.747  -30.346]
     [ -73.136  -43.425]
     [ -64.882  -43.255]
     [ -59.509  -25.698]
     [ -77.989   -8.823]
     [ 110.784    8.079]
     [  55.244 -124.371]
     [ -57.983  -28.766]
     [ -81.834   19.125]
     [-124.057   13.401]
     [  67.931   25.218]
     [-143.952  131.297]
     [ -70.100  160.068]
     [ -69.484  145.669]
     [ -77.264  124.223]
     [ -78.100      nan]]
    """
    bb_filter = filter_backbone(atom_array)
    backbone = atom_array[..., bb_filter]
    
    if backbone.array_length() % 3 != 0 \
        or (backbone.atom_name[0::3] != "N" ).any() \
        or (backbone.atom_name[1::3] != "CA").any() \
        or (backbone.atom_name[2::3] != "C" ).any():
            raise BadStructureError(
                "The backbone is invalid, must be repeats of (N, CA, C), "
                "maybe a backbone atom is missing"
            )
    phis = []
    psis = []
    omegas = []
    for chain_bb in chain_iter(backbone):
        phi, psi, omega = _dihedral_backbone(chain_bb)
        phis.append(phi)
        psis.append(psi)
        omegas.append(omega)
    return np.concatenate(phis, axis=-1), np.concatenate(psis, axis=-1), \
        np.concatenate(omegas, axis=-1)



def _dihedral_backbone(chain_bb):
    bb_coord = chain_bb.coord
    # Coordinates for dihedral angle calculation
    # Dim 0: Model index (only for atom array stacks)
    # Dim 1: Angle index
    # Dim 2: X, Y, Z coordinates
    # Dim 3: Atoms involved in dihedral angle
    if isinstance(chain_bb, AtomArray):
        angle_coord_shape = (len(bb_coord)//3, 3, 4)
    elif isinstance(chain_bb, AtomArrayStack):
        angle_coord_shape = (bb_coord.shape[0], bb_coord.shape[1]//3, 3, 4)
    phi_coord   = np.full(angle_coord_shape, np.nan)
    psi_coord   = np.full(angle_coord_shape, np.nan)
    omega_coord = np.full(angle_coord_shape, np.nan)
    
    # Indices for coordinates of CA atoms 
    ca_i = np.arange(bb_coord.shape[-2]//3) * 3 + 1
    phi_coord  [..., 1: , :, 0]  = bb_coord[..., ca_i[1: ]-2 ,:]
    phi_coord  [..., 1: , :, 1]  = bb_coord[..., ca_i[1: ]-1 ,:]
    phi_coord  [..., 1: , :, 2]  = bb_coord[..., ca_i[1: ]   ,:]
    phi_coord  [..., 1: , :, 3]  = bb_coord[..., ca_i[1: ]+1 ,:]
    psi_coord  [..., :-1, :, 0]  = bb_coord[..., ca_i[:-1]-1 ,:]
    psi_coord  [..., :-1, :, 1]  = bb_coord[..., ca_i[:-1]   ,:]
    psi_coord  [..., :-1, :, 2]  = bb_coord[..., ca_i[:-1]+1 ,:]
    psi_coord  [..., :-1, :, 3]  = bb_coord[..., ca_i[:-1]+2 ,:]
    omega_coord[..., :-1, :, 0]  = bb_coord[..., ca_i[:-1]   ,:]
    omega_coord[..., :-1, :, 1]  = bb_coord[..., ca_i[:-1]+1 ,:]
    omega_coord[..., :-1, :, 2]  = bb_coord[..., ca_i[:-1]+2 ,:]
    omega_coord[..., :-1, :, 3]  = bb_coord[..., ca_i[:-1]+3 ,:]
    
    phi = dihedral(phi_coord[...,0], phi_coord[...,1],
                   phi_coord[...,2], phi_coord[...,3])
    psi = dihedral(psi_coord[...,0], psi_coord[...,1],
                   psi_coord[...,2], psi_coord[...,3])
    omega = dihedral(omega_coord[...,0], omega_coord[...,1],
                     omega_coord[...,2], omega_coord[...,3])
    
    return phi, psi, omega


def centroid(atoms):
    """
    Measure the centroid of a structure.
    
    Parameters
    ----------
    atoms: ndarray or AtomArray or AtomArrayStack
        The structures to determine the centroid from.
        Alternatively an ndarray containing the coordinates can be
        provided.
    
    Returns
    -------
    centroid : float or ndarray
        The centroid of the structure(s). :class:`ndarray` is returned when
        an :class:`AtomArrayStack` is given (centroid for each model).
    """
    return np.mean(coord(atoms), axis=-2)


def _call_non_index_function(function, expected_amount,
                             atoms, indices, periodic=False, box=None):
    """
    Call an `xxx()` function based on the parameters given to a
    `index_xxx()` function.
    """
    if indices.shape[-1] != expected_amount:
        raise ValueError(
            f"Expected length {expected_amount} in the last dimension "
            f"of the indices, but got length {indices.shape[-1]}"
        )
    coord_list = []
    for i in range(expected_amount):
        coord_list.append(coord(atoms)[..., indices[:,i], :])
    if periodic:
        if box is None:
            if isinstance(atoms, (AtomArray, AtomArrayStack)):
                box = atoms.box
            else:
                raise ValueError(
                    "If `atoms` are coordinates, "
                    "the box must be set explicitly"
                )
    else:
        box = None
    return function(*coord_list, box)


def _displacement_orthogonal_box(fractions, box, disp):
    """
    Fill in the PBC-aware displacement vector for non-PBC-aware
    displacements given as fractions of given box vectors.
    """
    # Fraction components are guaranteed to be positive 
    # Use fraction vector components with lower absolute
    # -> new_vec[i] = vec[i] - 1 if vec[i] > 0.5 else vec[i]
    fractions[fractions > 0.5] -= 1
    disp[:] = fraction_to_coord(fractions, box)


def _displacement_triclinic_box(fractions, box, disp):
    """
    Fill in the PBC-aware displacement vector for non-PBC-aware
    displacements given as fractions of given box vectors.
    """
    diffs = fraction_to_coord(fractions, box)
    # Fraction components are guaranteed to be positive 
    # Test all 3 fraction vector components
    # with positive and negative sign
    # (i,j,k in {-1, 0})
    # Hence, 8 periodic copies are tested
    periodic_shift = []
    for i in range(-1, 1):
        for j in range(-1, 1):
            for k in range(-1, 1):
                x = i*box[0,0] + j*box[1,0] + k*box[2,0]
                y = i*box[0,1] + j*box[1,1] + k*box[2,1]
                z = i*box[0,2] + j*box[1,2] + k*box[2,2]
                periodic_shift.append([x,y,z])
    periodic_shift = np.array(periodic_shift, dtype=disp.dtype)
    # Create 8 periodically shifted variants for each atom
    shifted_diffs = diffs[:, np.newaxis, :] + periodic_shift[np.newaxis, :, :]
    # Find for each atom the periodically shifted variant with lowest
    # distance
    # Lowest squared distance -> lowest distance
    sq_distance = vector_dot(shifted_diffs, shifted_diffs)
    # for each given non-PBC-aware displacement find the PBC-aware
    # displacement with the lowest distance
    disp[:] = shifted_diffs[
        np.arange(len(shifted_diffs)),
        np.argmin(sq_distance, axis=1)
    ]



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.structure.graphics"
__author__ = "Tom David Müller"
__all__ = ["plot_nucleotide_secondary_structure"]

import shutil
import numpy as np
from itertools import repeat
from .. import pseudoknots
from ...application.viennarna import RNAplotApp


def plot_nucleotide_secondary_structure(
    axes, base_labels, base_pairs, length, 
    layout_type=RNAplotApp.Layout.NAVIEW, draw_pseudoknots=True, 
    pseudoknot_order=None, angle=0, bond_linewidth=1, bond_linestyle=None, 
    bond_color='black', backbone_linewidth=1, backbone_linestyle='solid', 
    backbone_color='grey', base_text=None, base_box=None, 
    annotation_positions=None, annotation_offset=8.5, annotation_text=None, 
    border=0.03, bin_path="RNAplot"
    ):
    """
    Generate 2D plots of nucleic acid secondary structures using the
    interface to *RNAplot*, which is part of the *ViennaRNA* software
    package.

    Internally a :class:`biotite.application.viennarna.RNAplotApp` 
    instance is created to generate coordinates for each individual base 
    on a 2D plane. *ViennaRNA* must be installed in order to use this 
    function.

    Parameters
    ----------
    axes : Axes
        A *Matplotlib* axes, that is used as plotting area.
    base_labels : iterable
        The labels denoting the type of each base.
    base_pairs : ndarray, shape=(n,2)
        Each row corresponds to the positions of the bases in the
        sequence. The positions are counted from zero.
    length : int
        The number of bases in the sequence.
    layout_type : RNAplotApp.Layout, optional (default: RNAplotApp.Layout.NAVIEW)
        The layout type according to the *RNAplot* documentation.
    draw_pseudoknots : bool, optional (default: True)
        Whether pseudoknotted bonds should be drawn.
    pseudoknot_order : iterable, optional (default: None)
        The pseudoknot order of each pair in the input `base_pairs`.
        If no pseudoknot order is given, a solution determined by 
        :func:`biotite.structure.pseudoknots` is picked at random.
    angle : int or float, optional (default: 0)
        The angle the plot should be rotated.
    bond_linewidth : float or int or iterable, optional (default: 1)
        The linewidth of each bond. Provide a single value to set the
        linewidth for all bonds or an iterable to set the linewidth for
        each individual bond.
    bond_linestyle : str or iterable, optional (default: None)
        The *Matplotlib* compatible linestyle of each bond. Provide a
        single value to set the linewidth for all bonds or an iterable
        to set the linewidth for each individual bond. By default, solid
        lines are used for non-pseudoknotted bonds and dashed lines are
        used for pseudoknotted bonds.
    bond_color : str or ndarray, shape(n,) or shape(n,3) or shape(n,4), optional (default: 'black')
        The *Matplotlib* compatible color of each bond. Provide a single
        string to set the color for all bonds or an array to set the
        color for each individual bond.
    backbone_linewidth : float, optional (default: 1)
        The linewidth of the backbone.
    backbone_linestyle : str, optional (default: 'solid')
        The *Matplotlib* compatible linestyle of the backbone.
    backbone_color : str or ndarray, shape=(3,) or shape=(4,), dtype=float, optional (default: 'grey')
        The *Matplotlib* compatible color of the backbone.
    base_text : dict or iterable, optional (default: {'size': 'small'})
        The keyword parameters for the *Matplotlib* ``Text`` objects 
        denoting the type of each base. Provide a single value to set 
        the parameters for all labels or an iterable to set the 
        parameters for each individual label.
    base_box : dict or iterable, optional (default: {'pad'=0, 'color'='white'})
        The *Matplotlib* compatible properties of the ``FancyBboxPatch``
        surrounding the base labels. Provide a single dictionary to
        set the properties of all base lables or an iterable to set the
        properties for each individual label.
    annotation_positions : iterable, optional (default: None)
        The positions of the bases to be numbered. By default every
        second base is annotated. Please note that while the positions
        in the sequence are counted from zero, they are displayed on the
        graph counted from one.
    annotation_offset : int or float, optional (default: 8.5)
        The offset of the annotations from the base labels.
    annotation_text : dict or iterable, optional (default: {'size': 'small'})
        The keyword parameters for the *Matplotlib* ``Text`` objects 
        annotating the sequence. Provide a single value to set the 
        parameters for all annotations or an iterable to set the 
        parameters for each individual annotation.
    border : float, optional (default: 0.03)
        The percentage of the coordinate range to be left as whitespace
        to create a border around the plot.
    bin_path : str, optional
        Path to the *RNAplot* binary.
    """

    # Check if RNAplot is installed
    if shutil.which(bin_path) is None:
        raise FileNotFoundError(
            'RNAplot is not installed at the specified location, unable to '
            'plot secondary structure.'
        )

    # Get the unknotted base pairs
    if pseudoknot_order is None:
        # Get a random unknotted structure retaining the maximum number
        # of base pairs.
        pseudoknot_order = pseudoknots(base_pairs, max_pseudoknot_order=0)[0]
    unknotted_base_pairs = base_pairs[pseudoknot_order == 0]

    # If `bond_linewidth` is a single value, extrapolate
    if isinstance(bond_linewidth, int) or isinstance(bond_linewidth, float):
        bond_linewidth = np.full(base_pairs.shape[0], bond_linewidth)

    # If `bond_color` is not an array, extrapolate
    if not isinstance(bond_color, np.ndarray):
        bond_color = np.full(base_pairs.shape[0], bond_color)

    # Set the default properties of the Matplotlib `bbox` surrounding
    # the base labels
    if base_box is None:
        base_box=np.full(length, {'pad': 0, 'color': 'white'})
    # if `base_box` is a dictionary, extrapolate
    elif isinstance(base_box, dict):
        base_box = np.full(length, base_box)

    # By default pseudoknotted bonds are denoted as dashed lines, while
    # unknotted bonds are denoted as solid lines
    if bond_linestyle is None:
        bond_linestyle = np.full(base_pairs.shape[0], 'solid', dtype='object')
        bond_linestyle[pseudoknot_order != 0] = 'dashed'
    # If `bond_linestyle` is a string, extrapolate
    elif isinstance(bond_linestyle, str):
        bond_linestyle = np.full(base_pairs.shape[0], bond_linestyle)

    # If pseudoknots are not to be drawn, remove pseudoknotted bonds,
    # regardless of the given linestyles
    if not draw_pseudoknots:
        bond_linestyle[pseudoknot_order != 0] = 'None'

    # Set the default properties of the base labels
    if base_text is None:
        base_text = np.full(length, {'size': 'small'})
    elif isinstance(base_text, dict):
        base_text = np.full(length, base_text)

    # If no specific annotation positions are given, annotate every
    # second base pair
    if annotation_positions is None:
        annotation_positions = range(0, length, 2)

    # Set the default font properties of the base annotations
    if annotation_text is None:
        annotation_text = repeat({'size': 'small'})
    elif isinstance(annotation_text, dict):
        annotation_text = repeat(annotation_text)

    # Get coordinates for secondary structure plot
    coordinates = RNAplotApp.compute_coordinates(
        base_pairs=unknotted_base_pairs,
        length=length,
        bin_path=bin_path,
        layout_type=layout_type
    )

    # Rotate Coordinates
    if angle != 0:
        angle = np.deg2rad(angle)
        rot_matrix = np.array(
            [[np.cos(angle), -np.sin(angle)],
            [np.sin(angle),  np.cos(angle)]]
        )
        for i, coord in enumerate(coordinates):
            coordinates[i] = np.dot(rot_matrix, coord)

    # Remove axes and frame
    axes.set_xticks([])
    axes.set_yticks([])
    axes.xaxis.set_tick_params(
        top=False, bottom=False, labeltop=False, labelbottom=False
    )
    axes.yaxis.set_tick_params(
        top=False, bottom=False, labeltop=False, labelbottom=False
    )
    axes.set_frame_on(False)


    # Define buffer area (Border)
    coord_range = abs(np.max(coordinates)) + abs(np.min(coordinates))
    buffer = border*coord_range

    # Adjust display
    axes.set_xlim(
        np.min(coordinates[:,0])-buffer, np.max(coordinates[:,0])+buffer
    )
    axes.set_ylim(
        np.min(coordinates[:,1])-buffer, np.max(coordinates[:,1])+buffer
    )
    axes.set_aspect(aspect='equal')

    # Draw backbone
    axes.plot(coordinates[:,0], coordinates[:,1], color=backbone_color,
              linestyle=backbone_linestyle, linewidth=backbone_linewidth)

    # Draw base labels
    for coords, label, box, text in zip(
        coordinates, base_labels, base_box, base_text
    ):
        t = axes.text(
                    x=coords[0], y=coords[1], s=label,
                    ha='center', va='center', **text
        )
        t.set_bbox(box)

    # Draw bonds
    for (base1, base2), color, style, width in zip(
        base_pairs, bond_color, bond_linestyle, bond_linewidth
    ):
        base1_coords = coordinates[base1]
        base2_coords = coordinates[base2]
        x = base1_coords[0], base2_coords[0]
        y = base1_coords[1], base2_coords[1]
        axes.plot(x, y, color=color, linestyle=style, linewidth=width)

    # Draw annotations
    for i, text in zip(annotation_positions, annotation_text):
        if (i > 0) and ((i+1) < length):
            # Get the average of the direction vectors to the next and
            # previous base
            vector_to_previous = np.array(
                [coordinates[i-1][0] - coordinates[i][0],
                 coordinates[i-1][1] - coordinates[i][1]]
            )
            vector_to_previous = vector_to_previous / np.linalg.norm(
                vector_to_previous
            )
            vector_to_next = np.array(
                [coordinates[i][0] - coordinates[i+1][0],
                 coordinates[i][1] - coordinates[i+1][1]]
            )
            vector_to_next = vector_to_next / np.linalg.norm(
                vector_to_next
            )
            vector = (vector_to_next + vector_to_previous) / 2
        elif i > 0:
            # For the last base get the direction vector to the previous
            # base
            vector = np.array(
                [coordinates[i-1][0] - coordinates[i][0],
                 coordinates[i-1][1] - coordinates[i][1]]
            )
        else:
            # For the first base get the direction vector to the next
            # base
            vector = np.array(
                [coordinates[i][0] - coordinates[i+1][0],
                 coordinates[i][1] - coordinates[i+1][1]]
            )
        # Normalize the vector
        vector = vector / np.linalg.norm(vector)
        # Get the perpendicular vector
        vector = np.array([vector[1], -vector[0]])
        # The annotations are offset in the direction of the
        # perpendicular vector
        x, y = coordinates[i] + (annotation_offset*vector)
        axes.text(
            x=x, y=y, s=i+1,
            ha='center', va='center', **text
        )


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

from tempfile import NamedTemporaryFile
import biotite
import biotite.sequence as seq
import biotite.sequence.io as seqio
import numpy as np
import glob
from os.path import join
from ..util import data_dir
import pytest


@pytest.mark.parametrize(
    "path", glob.glob(join(data_dir("sequence"), "random.*"))
)
def test_loading_single(path):
    ref_sequence = seqio.load_sequence(
        join(data_dir("sequence"), "random.fasta")
    )
    sequence = seqio.load_sequence(path)
    assert ref_sequence == sequence


@pytest.mark.parametrize("suffix", ["fasta", "fastq"])
def test_saving_single(suffix):
    ref_sequence = seqio.load_sequence(
        join(data_dir("sequence"), "random.fasta")
    )
    temp = NamedTemporaryFile("w+", suffix=f".{suffix}")
    try:
        seqio.save_sequence(temp.name, ref_sequence)
    except PermissionError:
        # This error might occur on AppVeyor
        pytest.skip("Permission is denied")


@pytest.mark.parametrize(
    "path", glob.glob(join(data_dir("sequence"), "random.*"))
)
def test_loading_multiple(path):
    ref_sequences = seqio.load_sequences(
        join(data_dir("sequence"), "random.fasta")
    )
    sequences = seqio.load_sequences(path)
    assert ref_sequences == sequences


@pytest.mark.parametrize("suffix", ["fasta", "fastq"])
def test_saving_multiple(suffix):
    ref_sequences = seqio.load_sequences(
        join(data_dir("sequence"), "random.fasta")
    )
    temp = NamedTemporaryFile("w+", suffix=f".{suffix}")
    try:
        seqio.save_sequences(temp.name, ref_sequences)
    except PermissionError:
        # This error might occur on AppVeyor
        pytest.skip("Permission is denied")

@pytest.mark.parametrize("file_name", ["gg_avidin.gb", "bt_lysozyme.gp"])
def test_genbank(file_name):
    """
    Simply test whether reading or writing a GenBank/GenPept file
    raises an exception.
    """

    temp = NamedTemporaryFile("w+", suffix=".gb")
    sequence = seqio.load_sequence(join(data_dir("sequence"), file_name))
    try:
        seqio.save_sequence(temp.name, sequence)
    except PermissionError:
        # This error might occur on AppVeyor
        pytest.skip("Permission is denied")


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import biotite.structure as struc
import biotite.structure.io as strucio
import numpy as np
from os.path import join
from ..util import data_dir
import pytest


def test_sse():
    array = strucio.load_structure(join(data_dir("structure"), "3o5r.mmtf"))
    sse = struc.annotate_sse(array, "A")
    sse_str = "".join(sse.tolist())
    assert sse_str == ("caaaaaacccccccccccccbbbbbccccccbbbbccccccccccccccc"
                       "ccccccccccccbbbbbbcccccccaaaaaaaaaccccccbbbbbccccc"
                       "ccccccccccccbbbbbbbccccccccc")


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import biotite.sequence as seq
from biotite.sequence import Location, Feature, Annotation, AnnotatedSequence
import biotite.sequence.io.genbank as gb
import numpy as np
from os.path import join
from ..util import data_dir
import pytest


def test_annotation_creation():
    feature1 = Feature("CDS", [seq.Location(1,2)], qual={"gene" : "test1"})
    feature2 = Feature("CDS", [seq.Location(3,4)], qual={"gene" : "test2"})
    feature_list = [feature1, feature2]
    annotation = Annotation(feature_list)
    for feature in annotation:
        assert feature.key in [f.key for f in feature_list]
        assert feature.qual["gene"] in [
            f.qual["gene"] for f in feature_list
        ]

def test_annotation_concatenation():
    feature1 = Feature("CDS", [seq.Location(1,1)], qual={"gene" : "test1"})
    feature2 = Feature("CDS", [seq.Location(2,2)], qual={"gene" : "test2"})
    annot1 = Annotation([feature1, feature2])
    feature3 = Feature("CDS", [seq.Location(3,3)], qual={"gene" : "test3"})
    feature4 = Feature("CDS", [seq.Location(4,4)], qual={"gene" : "test4"})
    annot2 = Annotation([feature3, feature4])
    feature5 = Feature("CDS", [seq.Location(5,5)], qual={"gene" : "test5"})
    concat = annot1 + annot2 + feature5
    assert set([f.qual["gene"] for f in concat]) \
        == set(["test1", "test2", "test3", "test4", "test5"])

def test_annotation_indexing():
    feature1 = Feature("CDS", [Location(-10,30 )], qual={"gene" : "test1"})
    feature2 = Feature("CDS", [Location(20, 50 )], qual={"gene" : "test2"})
    feature3 = Feature("CDS", [Location(100,130)], qual={"gene" : "test3"})
    feature4 = Feature("CDS", [Location(150,250)], qual={"gene" : "test4"})
    feature5 = Feature("CDS", [Location(-50,200)], qual={"gene" : "test5"})
    annotation = Annotation([feature1,feature2,feature3,feature4,feature5])
    sub_annot = annotation[40:150]
    # Only one location per feature
    assert set([list(f.locs)[0].defect for f in sub_annot]) \
        == set([Location.Defect.MISS_LEFT, Location.Defect.NONE,
                (Location.Defect.MISS_LEFT | Location.Defect.MISS_RIGHT)])
    assert set([f.qual["gene"] for f in sub_annot]) \
        == set(["test2", "test3", "test5"])

def test_annotated_sequence():
    sequence = seq.NucleotideSequence("ATGGCGTACGATTAGAAAAAAA")
    feature1 = Feature("misc_feature", [Location(1,2), Location(11,12)],
                       {"note" : "walker"})
    feature2 = Feature("misc_feature", [Location(16,22)], {"note" : "poly-A"})
    annotation = Annotation([feature1, feature2])
    annot_seq = AnnotatedSequence(annotation, sequence)
    assert annot_seq[2] == "T"
    assert annot_seq.sequence[2] == "G"
    annot_seq2 = annot_seq[:16]
    assert annot_seq2.sequence == seq.NucleotideSequence("ATGGCGTACGATTAG")
    assert annot_seq[feature1] == seq.NucleotideSequence("ATAT")
    assert annot_seq[feature2] == seq.NucleotideSequence("AAAAAAA")
    annot_seq[feature1] = seq.NucleotideSequence("CCCC")
    assert annot_seq.sequence == seq.NucleotideSequence("CCGGCGTACGCCTAGAAAAAAA")

def test_reverse_complement():
    gb_file = gb.GenBankFile.read(join(data_dir("sequence"), "ec_bl21.gb"))
    annot_seq = gb.get_annotated_sequence(gb_file)
    assert annot_seq == annot_seq.reverse_complement().reverse_complement()


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import warnings
import pytest
import numpy as np
from biotite.structure import Atom
from biotite.structure import array
from biotite.structure import BondList
from biotite.structure import partial_charges


# Test the partial charge of carbon in the molecules given in table
# 3 of the Gasteiger-Marsili publication
# Since some of the molecules are not available in the Chemical 
# Components Dictionary, the respective AtomArrays are constructed via
# Biotite and the coordinates are arbitrarily set to the origin since
# the relevant information is the BondList

# Creating atoms to build molecules with
carbon   = Atom([0, 0, 0], element="C")
hydrogen = Atom([0, 0, 0], element="H")
oxygen   = Atom([0, 0, 0], element="O")
nitrogen = Atom([0, 0, 0], element="N")
fluorine = Atom([0, 0, 0], element="F")
sulfur   = Atom([0, 0, 0], element="S")


# Building molecules
methane = array([carbon, hydrogen, hydrogen, hydrogen, hydrogen])
methane.bonds = BondList(
    methane.array_length(),
    np.array([[0,1,1], [0,2,1], [0,3,1], [0,4,1]])
)
mol_length = methane.array_length()
methane.charge = np.array([0] * mol_length)


ethane = array(
    [carbon, carbon, hydrogen, hydrogen, hydrogen, hydrogen, hydrogen,
    hydrogen]
)
ethane.bonds = BondList(
    ethane.array_length(),
    np.array([
        [0,1,1], [0,2,1], [0,3,1], [0,4,1], [1,5,1], [1,6,1], [1,7,1]
    ])
)
mol_length = ethane.array_length()
ethane.charge = np.array([0] * mol_length)


ethylene = array(
    [carbon, carbon, hydrogen, hydrogen, hydrogen, hydrogen]
)
ethylene.bonds = BondList(
    ethylene.array_length(),
    np.array([[0,1,2], [0,2,1], [0,3,1], [1,4,1], [1,5,1]])
)
mol_length = ethylene.array_length()
ethylene.charge = np.array([0] * mol_length)


acetylene = array(
    [carbon, carbon, hydrogen, hydrogen]
)
acetylene.bonds = BondList(
    acetylene.array_length(),
    np.array([[0,1,3], [0,2,1], [1,3,1]])
)
mol_length = acetylene.array_length()
acetylene.charge = np.array([0] * mol_length)


fluoromethane = array(
    [carbon, fluorine, hydrogen, hydrogen, hydrogen]
)
fluoromethane.bonds = BondList(
    fluoromethane.array_length(),
    np.array([[0,1,1], [0,2,1], [0,3,1], [0,4,1]])
)
mol_length = fluoromethane.array_length()
fluoromethane.charge = np.array([0] * mol_length)


difluoromethane = array(
    [carbon, fluorine, fluorine, hydrogen, hydrogen]
)
difluoromethane.bonds = BondList(
    difluoromethane.array_length(),
    np.array([[0,1,1], [0,2,1], [0,3,1], [0,4,1]])
)
mol_length = difluoromethane.array_length()
difluoromethane.charge = np.array([0] * mol_length)


trifluoromethane = array(
    [carbon, fluorine, fluorine, fluorine, hydrogen]
)
trifluoromethane.bonds = BondList(
    trifluoromethane.array_length(),
    np.array([[0,1,1], [0,2,1], [0,3,1], [0,4,1]])
)
mol_length = trifluoromethane.array_length()
trifluoromethane.charge = np.array([0] * mol_length)


tetrafluoromethane = array(
    [carbon, fluorine, fluorine, fluorine, fluorine]
)
tetrafluoromethane.bonds = BondList(
    tetrafluoromethane.array_length(),
    np.array([[0,1,1], [0,2,1], [0,3,1], [0,4,1]])
)
mol_length = tetrafluoromethane.array_length()
tetrafluoromethane.charge = np.array([0] * mol_length)


fluoroethane = array(
    [carbon, carbon, fluorine, hydrogen, hydrogen, hydrogen, 
    hydrogen, hydrogen]
)
fluoroethane.bonds = BondList(
    fluoroethane.array_length(),
    np.array([
        [0,1,1], [0,2,1], [0,3,1], [0,4,1], [1,5,1], [1,6,1], [1,7,1]
    ])
)
mol_length = fluoroethane.array_length()
fluoroethane.charge = np.array([0] * mol_length)


trifluoroethane = array(
    [carbon, carbon, fluorine, fluorine, fluorine, hydrogen,
    hydrogen, hydrogen]
)
trifluoroethane.bonds = BondList(
    trifluoroethane.array_length(),
    np.array([
        [0,1,1], [0,2,1], [0,3,1], [0,4,1], [1,5,1], [1,6,1], [1,7,1]
    ])
)
mol_length = trifluoroethane.array_length()
trifluoroethane.charge = np.array([0] * mol_length)


methanole = array(
    [carbon, oxygen, hydrogen, hydrogen, hydrogen, hydrogen]
)
methanole.bonds = BondList(
    methanole.array_length(),
    np.array([[0,1,1], [0,2,1], [0,3,1], [0,4,1], [1,5,1]])
)
mol_length = methanole.array_length()
methanole.charge = np.array([0] * mol_length)


dimethyl_ether = array(
    [carbon, carbon, oxygen, hydrogen, hydrogen, hydrogen, hydrogen,
    hydrogen, hydrogen]
)
dimethyl_ether.bonds = BondList(
    dimethyl_ether.array_length(),
    np.array([
        [0,2,1], [1,2,1], [0,3,1], [0,4,1], [0,5,1], [1,6,1], [1,7,1],
        [1,8,1]
    ])
)
mol_length = dimethyl_ether.array_length()
dimethyl_ether.charge = np.array([0] * mol_length)


formaldehyde = array(
    [carbon, oxygen, hydrogen, hydrogen]
)
formaldehyde.bonds = BondList(
    formaldehyde.array_length(),
    np.array([[0,1,2], [0,2,1], [0,3,1]])
)
mol_length = formaldehyde.array_length()
formaldehyde.charge = np.array([0] * mol_length)


acetaldehyde = array(
    [carbon, carbon, oxygen, hydrogen, hydrogen, hydrogen, hydrogen]
)
acetaldehyde.bonds = BondList(
    acetaldehyde.array_length(),
    np.array([[0,1,1], [1,2,2], [0,3,1], [0,4,1], [0,5,1], [1,6,1]])
)
mol_length = acetaldehyde.array_length()
acetaldehyde.charge = np.array([0] * mol_length)


acetone = array(
    [carbon, carbon, carbon, oxygen, hydrogen, hydrogen, hydrogen,
    hydrogen, hydrogen, hydrogen]
)
acetone.bonds = BondList(
    acetone.array_length(),
    np.array([
        [0,1,1], [1,2,1], [1,3,2], [0,4,1], [0,5,1], [0,6,1], [2,7,1],
        [2,8,1], [2,9,1]
    ])
)
mol_length = acetone.array_length()
acetone.charge = np.array([0] * mol_length)


hydrogen_cyanide = array(
    [carbon, nitrogen, hydrogen]
)
hydrogen_cyanide.bonds = BondList(
    hydrogen_cyanide.array_length(),
    np.array([[0,1,3], [0,2,1]])
)
mol_length = hydrogen_cyanide.array_length()
hydrogen_cyanide.charge = np.array([0] * mol_length)


acetonitrile = array(
    [carbon, carbon, nitrogen, hydrogen, hydrogen, hydrogen]
)
acetonitrile.bonds = BondList(
    acetonitrile.array_length(),
    np.array([[0,1,1], [1,2,3], [0,3,1], [0,4,1], [0,5,1]])
)
mol_length = acetonitrile.array_length()
acetonitrile.charge = np.array([0] * mol_length)

# For this purpose, parametrization via pytest is performed
@pytest.mark.parametrize("molecule, expected_results", [
    (methane, (-0.078,)),
    (ethane, (-0.068, -0.068)),
    (ethylene, (-0.106, -0.106)),
    (acetylene, (-0.122, -0.122)),
    (fluoromethane, (0.079,)),
    (difluoromethane, (0.23,)),
    (trifluoromethane, (0.38,)),
    (tetrafluoromethane, (0.561,)),
    (fluoroethane, (0.087, -0.037)),
    (trifluoroethane, (0.387, 0.039)),
    (methanole, (0.033,)),
    (dimethyl_ether, (0.036, 0.036)),
    (formaldehyde, (0.115,)),
    (acetaldehyde, (-0.009, 0.123)),
    (acetone, (-0.006, 0.131, -0.006)),
    (hydrogen_cyanide, (0.051,)),
    (acetonitrile, (0.023, 0.06))
])
def test_partial_charges(molecule, expected_results):
    """
    Test whether the partial charges of the carbon atoms comprised in
    the molecules given in table 3 of the publication computed in this
    implementation correspond to the values given in the publication
    within a certain tolerance range.
    """
    charges = partial_charges(molecule)
    assert charges[molecule.element == "C"].tolist() == \
        pytest.approx(expected_results, abs=1e-2)


@pytest.mark.parametrize("molecule", [
    methane,
    ethane,
    ethylene,
    acetylene,
    fluoromethane,
    difluoromethane,
    trifluoromethane,
    tetrafluoromethane,
    fluoroethane,
    trifluoroethane,
    methanole,
    dimethyl_ether,
    formaldehyde,
    acetaldehyde,
    acetone,
    hydrogen_cyanide,
    acetonitrile
])
def test_total_charge_zero(molecule):
    """
    In the case of the 17 molecules given in table 3, it is verified
    whether the sum of all partial charges equals the sum
    of all formal charges (in our case zero since we are exclusively
    dealing with uncharged molecules).
    """
    total_charge = np.sum(partial_charges(molecule))
    assert total_charge == pytest.approx(0, abs=1e-6)


def test_pos_formal_charge():
    """
    Test whether the partial charge of carbon in methane behaves as
    expected if it carries a formal charge of +1. To be more precise,
    it is expected to be smaller than 1 since this is the value which
    negative charge is addded to during iteration and also greater than
    the partial charge of carbon in methane carrying no formal charge.
    """
    pos_methane = methane.copy()
    pos_methane.charge = np.array([1, 0, 0, 0, 0])

    ref_carb_part_charge = partial_charges(
        methane,
        iteration_step_num=6
    )[0]
    pos_carb_part_charge = partial_charges(
        pos_methane,
        iteration_step_num=6
    )[0]
    assert pos_carb_part_charge < 1
    assert pos_carb_part_charge > ref_carb_part_charge


def test_valence_state_not_parametrized():
    """
    Test case in which parameters for a certain valence state of a
    generally parametrized atom are not available.
    In our case, it is sulfur having a double bond, i. e. only one
    binding partner.
    For this purpose, the molecule thioformaldehyde consisting of a
    central carbon bound to two hydrogen atoms via single bonds and to
    one sulfur atom via a double bond is created and tested.
    The expectations are the following: the sulfur's partial charge to
    be NaN and the carbons's partial charge to be smaller than that of
    the two hydrogens. Furthermore, the respective warning is expected
    to be raised.
    """
    warning_message_no_param_for_valence = (
        "Parameters for specific valence states of some atoms are not "
        "available. These valence states are: \n"
        "Atom:     Amount of binding partners:     Bond type:\n"
        "S         ---------------------------     2\n"
        "Their electronegativity is given as NaN."
    )
    with pytest.warns(None) as record:
        thioformaldehyde = array(
            [carbon, sulfur, hydrogen, hydrogen]
        )
        thioformaldehyde.bonds = BondList(
            thioformaldehyde.array_length(),
            np.array([[0,1,2], [0,2,1], [0,3,1]])
        )
        mol_length = thioformaldehyde.array_length()
        thioformaldehyde.charge = np.array([0] * mol_length)
        charges = partial_charges(thioformaldehyde)
        sulfur_part_charge = charges[1]
        carb_part_charge = charges[0]
        hyd_part_charge = charges[2]
    assert len(record) == 1
    warning = record[0]
    assert issubclass(warning.category, UserWarning)
    assert str(warning.message) == warning_message_no_param_for_valence
    assert np.isnan(sulfur_part_charge)
    assert carb_part_charge < hyd_part_charge


def test_correct_output_ions():
    """
    Ions such as sodium or potassium are not parametrized. However,
    their formal charge is taken as partial charge since they are not
    involved in covalent bonding.
    Hence, it is expected that no warning is raised.
    The test is performed with a sodium ion.
    """
    sodium = Atom([0, 0, 0], element="NA")
    sodium_array = array([sodium])
    # Sodium possesses a formal charge of +1
    sodium_array.charge = np.array([1])
    # Sodium is not involved in covalent bonding
    sodium_array.bonds = BondList(sodium_array.array_length())
    with pytest.warns(None) as record:
        sodium_charge = partial_charges(
            sodium_array, iteration_step_num=1
        )[0]
    assert len(record) == 0
    assert sodium_charge == 1


def test_correct_output_charged_aa():
    """
    Identification of hybridisation state is primarily performed via the
    bond type; if the bond type is BondType.ANY, it is resorted to the
    amount of binding partners of the respective atom, which can lead to
    an erroneous identification if the respective atom possesses a
    formal charge due to one additional binding partner or the loss of
    one binding partner. As an example, oxygen of an hydroxyl group has
    an formal charge of -1 if it loses the hydrogen, but still has the
    same hybridisation state. Identification via the amount of binding
    partners would be wrong.

    Therefore, the aim of this test is to verify that a molecule, in our
    case the amino acid glycine, with charges due to the addition or the
    loss of protons, delivers values for the partial charges of the
    atoms where addition/loss of the protons has taken place in an
    expected manner.
    A glycine AtomArray is constructed such that the carboxyl group is
    deprotonated and the amino group is protonated.

    Precisely, in the case of the negatively charged oxygen atom in the
    carboxyl group, a lower electronegativity value and therefore a
    lower partial charge is expected in case of correct identification
    of the hybridisation state via the bond type than in case of an
    erroneous identification via the amount of binding partners (cf.
    equation 2 for electronegativity and parameters listed in table 1 of
    the respective publication of Gasteiger and Marsili). The case of
    the positively charged nitrogen in the amino group, a correct
    identification of the hybridisation state via the amount of binding
    partners is possible as an amount of four binding partners is
    unambiguous, i. e. only represents the sp3 hybridisation. Hence,
    it is verified whether both ways of the identification of the
    hybridisation state deliver approximately the same result (small
    deviation arises from the difference in the parameters for the
    oxygen atom of the hydroxyl group in the carboxyl group that
    propagates to the amino group).
    Moreover, it is verified whether the respective UserWarning about
    unspecified bond types throughout the whole AtomArray is raised.
    """

    warning_message_unspecified_btype_throughout_the_array = (
        "Each atom's bond type is 0 (any). Therefore, it is resorted "
        "to the amount of binding partners for the identification of "
        "the hybridisation state which can lead to erroneous results."
    )

    glycine_charge = np.array(
        [+1, 0, 0, 0, -1, 0, 0, 0, 0, 0]
    )

    glycine_with_btype = array(
        [nitrogen, carbon, carbon, oxygen, oxygen, hydrogen, hydrogen,
            hydrogen, hydrogen, hydrogen]
    )
    glycine_with_btype.charge = glycine_charge
    glycine_with_btype.bonds = BondList(
        glycine_with_btype.array_length(),
        np.array([
            [0,1,1], [0,5,1], [0,6,1], [0,7,1], [1,2,1], [1,8,1],
            [1,9,1], [2,3,2], [2,4,1]
        ])
    )

    glycine_without_btype = glycine_with_btype.copy()
    glycine_without_btype.charge = glycine_charge
    glycine_without_btype.bonds = BondList(
        glycine_without_btype.array_length(),
        np.array([
            [0,1,0], [0,5,0], [0,6,0], [0,7,0], [1,2,0], [1,8,0],
            [1,9,0], [2,3,0], [2,4,0]
        ])
    )

    part_charges_with_btype = partial_charges(glycine_with_btype)
    with pytest.warns(None) as record:
        part_charges_without_btype = partial_charges(
            glycine_without_btype
        )
    assert len(record) == 1
    warning = record[0]
    assert issubclass(warning.category, UserWarning)
    assert str(warning.message) == \
        warning_message_unspecified_btype_throughout_the_array

    # Nitrogen of the amino group has the index 0
    nitr_charge_with_btype = part_charges_with_btype[0]
    nitr_charge_without_btype = part_charges_without_btype[0]
    assert nitr_charge_with_btype == pytest.approx(
        nitr_charge_without_btype, abs=5e-4
    )

    # Oxygen of the hydroxyl group in the carboxyl group has the index 2
    oxyg_charge_with_btype = part_charges_with_btype[2]
    oxyg_charge_without_btype = part_charges_without_btype[2]
    assert oxyg_charge_with_btype < oxyg_charge_without_btype
    # Assert that difference between the two values is significant
    difference_oxyg_charges = abs(oxyg_charge_with_btype
        - oxyg_charge_without_btype)
    assert difference_oxyg_charges > 3e-2


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.structure.info"
__author__ = "Patrick Kunzmann"
__all__ = ["mass"]

import json
from os.path import join, dirname, realpath
import msgpack
from ..atoms import Atom, AtomArray, AtomArrayStack


_info_dir = dirname(realpath(__file__))
# Masses are taken from http://www.sbcs.qmul.ac.uk/iupac/AtWt/ (2018/03/01)
with open(join(_info_dir, "atom_masses.json"), "r") as file:
    _atom_masses = json.load(file)
# Masses are taken from
# ftp://ftp.wwpdb.org/pub/pdb/data/monomers/components.cif
# (2019/01/27)
with open(join(_info_dir, "residue_masses.msgpack"), "rb") as file:
    _res_masses = msgpack.load(file, raw=False)


def mass(item, is_residue=None):
    """
    Calculate the mass for the given object.
    :footcite:`Meija2016`

    If a residue name is given, the mass values refer to the masses of
    the complete molecule without additional or missing protons.
    In case of residues in a longer chain, some atoms might be missing
    from the molecule.
    For example non-terminal residues in a protein or nucleotide chain
    miss the mass of a water molecule.
    
    Parameters
    ----------
    item : str or Atom or AtomArray or AtomArrayStack
        The atom or molecule to get the mass for.
        If a string is given, it is interpreted as residue name or
        chemical element.
        If an :class:`Atom` is given the mass is taken from its element.
        If an :class:`AtomArray` or :class:`AtomArrayStack` is given the
        mass is the sum of the mass of its atoms.
    is_residue : bool, optional
        If set to true and a string is given for `item`, the string
        will be strictly interpreted as residue.
        If set to false, the string is strictly interpreted as element.
        By default the string will be interpreted as element at first
        and secondly as residue name, if the element is unknown.
    
    Returns
    -------
    mass : float or None
        The mass of the given object in *u*. None if the mass is unknown.
    
    References
    ----------
    
    .. footbibliography::
    
    Examples
    --------

    >>> print(mass(atom_array))
    2170.438
    >>> first_residue = list(residue_iter(atom_array))[0]
    >>> print(first_residue)
        A       1  ASN N      N        -8.901    4.127   -0.555
        A       1  ASN CA     C        -8.608    3.135   -1.618
        A       1  ASN C      C        -7.117    2.964   -1.897
        A       1  ASN O      O        -6.634    1.849   -1.758
        A       1  ASN CB     C        -9.437    3.396   -2.889
        A       1  ASN CG     C       -10.915    3.130   -2.611
        A       1  ASN OD1    O       -11.269    2.700   -1.524
        A       1  ASN ND2    N       -11.806    3.406   -3.543
        A       1  ASN H1     H        -8.330    3.957    0.261
        A       1  ASN H2     H        -8.740    5.068   -0.889
        A       1  ASN H3     H        -9.877    4.041   -0.293
        A       1  ASN HA     H        -8.930    2.162   -1.239
        A       1  ASN HB2    H        -9.310    4.417   -3.193
        A       1  ASN HB3    H        -9.108    2.719   -3.679
        A       1  ASN HD21   H       -11.572    3.791   -4.444
        A       1  ASN HD22   H       -12.757    3.183   -3.294
    >>> print(mass("ASN"))
    132.118
    >>> first_atom = first_residue[0]
    >>> print(first_atom)
        A       1  ASN N      N        -8.901    4.127   -0.555
    >>> print(mass(first_atom))
    14.007
    >>> print(mass("N"))
    14.007
    """

    if isinstance(item, str):
        if is_residue is None:
            result_mass = _atom_masses.get(item.upper())
            if result_mass is None:
                result_mass = _res_masses.get(item.upper())
        elif not is_residue:
            result_mass = _atom_masses.get(item.upper())
        else:
            result_mass = _res_masses.get(item.upper())
    
    elif isinstance(item, Atom):
        result_mass = mass(item.element, is_residue=False)
    elif isinstance(item, AtomArray) or isinstance(item, AtomArrayStack):
        result_mass = sum(
            (mass(element, is_residue=False) for element in item.element)
        )
    
    else:
        raise TypeError(
            f"Cannot calculate mass for {type(item).__name__} objects"
        )
    
    if result_mass is None:
        raise KeyError(f"{item} is not known")
    return result_mass


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
This module provides functions to calculate the radial distribution function.
"""

__name__ = "biotite.structure"
__author__ = "Daniel Bauer, Patrick Kunzmann"
__all__ = ["rdf"]

from numbers import Integral
import numpy as np
from .atoms import Atom, AtomArray, stack, array, coord, AtomArrayStack
from .box import box_volume
from .geometry import displacement
from .util import vector_dot
from .celllist import CellList


def rdf(center, atoms, selection=None, interval=(0, 10), bins=100, box=None,
        periodic=False):
    r"""
    Compute the radial distribution function *g(r)* (RDF) for one or
    multiple given central positions based on a given system of
    particles.

    Parameters
    ----------
    center : Atom or AtomArray or AtomArrayStack or ndarray, dtype=float
        Coordinates or atoms(s) to use as origin(s) for RDF calculation.

        - If a single :class:`Atom` or a :class:`ndarray` with shape
          *(3,)* is given, the RDF is only calculated for this position.
        - If an :class:`AtomArray` or a :class:`ndarray` with shape
          *(n,3)* is given, the calculated RDF histogram is an average
          over *n* postions.
        - If an :class:`AtomArrayStack` or a :class:`ndarray` with shape
          *(m,n,3)* is given, different centers are used for each model
          *m*.
          The calculated RDF histogram is an average over *m*
          models and *n* positions.
          This requires `atoms` to be an :class:`AtomArrayStack`.

    atoms : AtomArray or AtomArrayStack
        The distribution is calculated based on these atoms.
        When an an :class:`AtomArrayStack` is provided, the RDF
        histogram is averaged over all models.
        Please not that `atoms` must have an associated box,
        unless `box` is set.
    selection : ndarray, dtype=bool, shape=(n,), optional
        Boolean mask for `atoms` to limit the RDF calculation to
        specific atoms.
    interval : tuple, optional
        The range in which the RDF is calculated.
    bins : int or sequence of scalars or str, optional
        Bins for the RDF.

        - If `bins` is an `int`, it defines the number of bins for the
          given `interval`.
        - If `bins` is a sequence, it defines the bin edges, ignoring
          the `interval` parameter. The output `bins` has the length
          of this input parameter reduced by one.
        - If `bins` is a string, it defines the function used to
          calculate the bins.

        See `numpy.histogram()` for further details.
    box : ndarray, shape=(3,3) or shape=(m,3,3), optional
        If this parameter is set, the given box is used instead of the
        `box` attribute of `atoms`.
        Must have shape *(3,3)* if atoms is an :class:`AtomArray` or
        *(m,3,3)* if atoms is an :class:`AtomArrayStack`, respectively.
    periodic : bool, optional
        Defines if periodic boundary conditions are taken into account.

    Returns
    -------
    bins : ndarray, dtype=float, shape=n
        The centers of the histogram bins.
        The length of the array is given by the `bins` input parameter.
    rdf : ndarry, dtype=float, shape=n
        RDF values for every bin.

    Notes
    -----
    Since the RDF depends on the average particle density of the system,
    this function strictly requires an box.

    Examples
    --------
    Calculate the oxygen-oxygen radial distribution function of water.
    The range of the histogram starts at 0.2 Å, in order to
    ignore the counts for the density for each oxygen to itself.

    >>> from os.path import join
    >>> waterbox = load_structure(join(path_to_structures, "waterbox.gro"))
    >>> oxygens = waterbox[:, waterbox.atom_name == 'OW']
    >>> bins, g_r = rdf(oxygens, oxygens, bins=49, interval=(0.2, 10), periodic=True)

    Print the RDF depending on the radius. Bins are in Å.

    >>> for x, y in zip(bins, g_r):
    ...     print(f"{x:.2f} {y:.2f}")
    0.30 0.00
    0.50 0.00
    0.70 0.04
    0.90 0.02
    1.10 0.03
    1.30 0.06
    1.50 0.03
    1.70 0.04
    1.90 0.04
    2.10 0.04
    2.30 0.04
    2.50 0.16
    2.70 1.99
    2.90 2.22
    3.10 1.34
    3.30 1.04
    3.50 0.97
    3.70 0.94
    3.90 0.97
    4.10 0.94
    4.30 0.98
    4.50 0.97
    4.70 0.96
    4.90 0.99
    5.10 0.99
    5.30 1.02
    5.50 1.02
    5.70 0.99
    5.90 0.98
    6.10 0.98
    6.30 0.99
    6.50 1.02
    6.70 1.02
    6.90 1.00
    7.10 1.01
    7.30 1.01
    7.50 1.00
    7.70 1.01
    7.90 0.99
    8.10 0.99
    8.30 0.99
    8.50 0.99
    8.70 0.99
    8.90 1.00
    9.10 1.01
    9.30 1.01
    9.50 1.00
    9.70 1.00
    9.90 0.99

    Find the radius for the first solvation shell.
    In this simple case, the density peak is identified by finding
    the maximum of the function.
    
    >>> peak_position = np.argmax(g_r)
    >>> print(f"{bins[peak_position]/10:.2f} nm")
    0.29 nm
    """
    if isinstance(atoms, AtomArray):
        # Reshape always to a stack for easier calculation
        atoms = stack([atoms])
    if selection is not None:
        atoms = atoms[..., selection]
    
    atom_coord = atoms.coord

    if box is None:
        if atoms.box is None:
            raise ValueError("A box must be supplied")
        else:
            box = atoms.box
    
    center = coord(center)
    if center.ndim == 1:
        center = center.reshape((1, 1) + center.shape)
    elif center.ndim == 2:
        center = center.reshape((1,) + center.shape)
    
    if box.shape[0] != center.shape[0] or box.shape[0] != atom_coord.shape[0]:
        raise ValueError(
            "Center, box, and atoms must have the same model count"
        )

    # Calculate distance histogram
    edges = _calculate_edges(interval, bins)
    # Make histogram of quared distances to save computation time
    # of sqrt calculation
    sq_edges = edges**2
    threshold_dist = edges[-1]
    cell_size = threshold_dist
    disp = []
    for i in range(atoms.stack_depth()):
        # Use cell list to efficiently preselect atoms that are in range
        # of the desired bin range
        cell_list = CellList(atom_coord[i], cell_size, periodic, box[i])
        # 'cell_radius=1' is used in 'get_atoms_in_cells()'
        # This is enough to find all atoms that are in the given
        # interval (and more), since the size of each cell is as large
        # as the last edge of the bins
        near_atom_mask = cell_list.get_atoms_in_cells(center[i], as_mask=True)
        # Calculate distances of each center to preselected atoms
        # for each center
        for j in range(center.shape[1]):
            dist_box = box[i] if periodic else None
            # Calculate squared distances
            disp.append(displacement(
                center[i,j], atom_coord[i, near_atom_mask[j]], box=dist_box
            ))
    # Make one array from multiple arrays with different length
    disp = np.concatenate(disp)
    sq_distances = vector_dot(disp, disp)
    hist, _ = np.histogram(sq_distances, bins=sq_edges)

    # Normalize with average particle density (N/V) in each bin
    bin_volume =   (4 / 3 * np.pi * np.power(edges[1: ], 3)) \
                 - (4 / 3 * np.pi * np.power(edges[:-1], 3))
    n_frames = len(atoms)
    volume = box_volume(box).mean()
    density = atoms.array_length() / volume
    g_r = hist / (bin_volume * density * n_frames)

    # Normalize with number of centers
    g_r /= center.shape[1]

    bin_centers = (edges[:-1] + edges[1:]) * 0.5

    return bin_centers, g_r


def _calculate_edges(interval, bins):
    if isinstance(bins, Integral):
        if bins < 1:
            raise ValueError("At least one bin is required")
        return np.linspace(*interval, bins+1)
    else:
        # 'bins' contains edges
        return np.array(bins, dtype=float)



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
This subpackage provides functionality for sequence alignments.

The two central classes involved are :class:`SubstitutionMatrix` and
:class:`Alignment`:

Every function that performs an alignment requires a
:class:`SubstitutionMatrix` that provides similarity scores for each
symbol combination of two alphabets (usually both alphabets are equal).
The alphabets in the :class:`SubstitutionMatrix` must match or extend
the alphabets of the sequences to be aligned.

An alignment cannot be directly represented as list of :class:`Sequence`
objects, since a gap indicates the absence of any symbol.
Instead, the aligning functions return one or more :class:`Alignment`
instances.
These objects contain the original sequences and a trace, that describe
which positions (indices) in the sequences are aligned.
Optionally they also contain the similarity score.

The aligning functions are usually C-accelerated, reducing the
computation time substantially.

This subpackage also contains functionality for finding *k-mer* matches
between two sequences, allowing fast heuristic pairwise alignments. 
"""

__name__ = "biotite.sequence.align"
__author__ = "Patrick Kunzmann"

from .alignment import *
from .banded import *
from .kmeralphabet import *
from .kmertable import *
from .kmersimilarity import *
from .localgapped import *
from .localungapped import *
from .matrix import *
from .multiple import *
from .pairwise import *
from .statistics import *


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

from os.path import join
import itertools
import numpy as np
import pytest
import biotite.structure as struc
from biotite.structure.io import load_structure
from ..util import data_dir, cannot_import


SAMPLE_BOXES = [
    (1, 1, 1,  90,  90,  90),
    (9, 5, 2,  90,  90,  90),
    (5, 5, 8,  90,  90, 120),
    (3, 2, 1,  10,  20,  20),
    (2, 4, 6, 100, 110, 120),
    (9, 9, 9,  90,  90, 170),
    (9, 8, 7,  50,  80,  50),
]

SAMPLE_COORD = [
    ( 1,  1,  1),
    ( 5, 10, 20),
    (-1,  5,  8),
    ( 3,  1, 54)
]



# Ignore warning about dummy unit cell vector
@pytest.mark.filterwarnings("ignore")
@pytest.mark.skipif(
    cannot_import("mdtraj"),
    reason="MDTraj is not installed"
)
@pytest.mark.parametrize(
    "len_a, len_b, len_c, alpha, beta, gamma", SAMPLE_BOXES
)
def test_box_vector_calculation(len_a, len_b, len_c, alpha, beta, gamma):
    box = struc.vectors_from_unitcell(
        len_a, len_b, len_c,
        np.deg2rad(alpha), np.deg2rad(beta), np.deg2rad(gamma)
    )

    from mdtraj.utils import lengths_and_angles_to_box_vectors
    ref_box = np.stack(
        lengths_and_angles_to_box_vectors(
            len_a, len_b, len_c, alpha, beta, gamma
        )
    )
    assert np.allclose(box, ref_box)

    assert struc.unitcell_from_vectors(box) == pytest.approx(
        (len_a, len_b, len_c,
         alpha * 2*np.pi / 360, beta * 2*np.pi / 360, gamma * 2*np.pi / 360)
    )


def test_volume():
    # Very rudimentary test
    box = np.array([
        [5,0,0],
        [0,8,0],
        [0,0,2],
    ])
    assert struc.box_volume(box) == pytest.approx(80)
    boxes = np.stack([box, box])
    assert struc.box_volume(boxes) == pytest.approx(80,80)


@pytest.mark.parametrize(
    "len_a, len_b, len_c, alpha, beta, gamma, x, y,z",
    [box+coord for box, coord in itertools.product(SAMPLE_BOXES, SAMPLE_COORD)]
)
def test_move_into_box(len_a, len_b, len_c, alpha, beta, gamma, x, y, z):
    box = struc.vectors_from_unitcell(
        len_a, len_b, len_c,
        np.deg2rad(alpha), np.deg2rad(beta), np.deg2rad(gamma)
    )
    coord = np.array([x,y,z])

    moved_coord = struc.move_inside_box(coord, box)
    fractions = struc.coord_to_fraction(moved_coord, box)
    assert ((fractions >= 0) & (fractions <=1)).all()


@pytest.mark.parametrize(
    "len_a, len_b, len_c, alpha, beta, gamma, x, y,z",
    [box+coord for box, coord in itertools.product(SAMPLE_BOXES, SAMPLE_COORD)]
)
def test_conversion_to_fraction(len_a, len_b, len_c,
                                alpha, beta, gamma,
                                x, y, z):
    box = struc.vectors_from_unitcell(
        len_a, len_b, len_c,
        np.deg2rad(alpha), np.deg2rad(beta), np.deg2rad(gamma)
    )
    coord = np.array([x,y,z])

    fractions = struc.coord_to_fraction(coord, box)
    if struc.is_orthogonal(box):
        assert fractions.tolist() == pytest.approx(coord / np.diagonal(box))
    new_coord = struc.fraction_to_coord(fractions, box)
    assert np.allclose(coord, new_coord)

    coords = np.stack([coord, coord])
    boxes = np.stack([box, box])
    fractions = struc.coord_to_fraction(coords, boxes)
    new_coords = struc.fraction_to_coord(fractions, boxes)
    assert np.allclose(coords, new_coords)


def test_repeat_box():
    array = load_structure(join(data_dir("structure"), "3o5r.mmtf"))
    repeat_array, _ = struc.repeat_box(array)
    assert repeat_array[:array.array_length()] == array


def test_remove_pbc_unsegmented():
    """
    `remove_pbc()` should not alter unsegmented structures,
    when the structure is entirely in the box.
    Exclude the solvent, due to high distances between each atom. 
    """
    ref_array = load_structure(join(data_dir("structure"), "3o5r.mmtf"))
    # Center structure in box
    centroid = struc.centroid(ref_array)
    box_center = np.diag(ref_array.box) / 2
    ref_array = struc.translate(ref_array, box_center-centroid)
    # Remove solvent
    ref_array = ref_array[~struc.filter_solvent(ref_array)]
    array = struc.remove_pbc(ref_array)

    assert ref_array.equal_annotation_categories(array)
    assert np.allclose(ref_array.coord, array.coord)


@pytest.mark.parametrize(
    "multi_model, translation_vector",
    itertools.product(
        [False, True],
        [(20,30,40), (-11, 33, 22), (-40, -50, -60)]
    )
)
def test_remove_pbc_restore(multi_model, translation_vector):
    CUTOFF = 5.0
    
    def get_matrices(array):
        """
        Create a periodic and non-periodic adjacency matrix.
        """
        nonlocal CUTOFF
        if isinstance(array, struc.AtomArray):
            matrix     = struc.CellList(array, CUTOFF, periodic=False) \
                        .create_adjacency_matrix(CUTOFF)
            matrix_pbc = struc.CellList(array, CUTOFF, periodic=True) \
                        .create_adjacency_matrix(CUTOFF)
        elif isinstance(array, struc.AtomArrayStack):
            matrix = np.array(
                [struc.CellList(model, CUTOFF, periodic=False)
                 .create_adjacency_matrix(CUTOFF)
                 for model in array]
                )
            matrix_pbc = np.array(
                [struc.CellList(model, CUTOFF, periodic=True)
                 .create_adjacency_matrix(CUTOFF)
                 for model in array]
            )
        return matrix, matrix_pbc
    
    def assert_equal_matrices(array, matrix1, matrix2, periodic):
        """
        Due to numerical instability, entries in both matrices might
        be different, when the distance of atoms is almost equal to
        the cutoff distance of the matrix.
        This function checks, whether two atoms with unequal entries
        in the matrices are near the cutoff distance.
        """
        nonlocal CUTOFF
        indices = np.where(matrix1 != matrix2)
        for index in range(len(indices[0])):
            if len(indices) == 2:
                # multi_model = False -> AtomArray
                m = None
                i = indices[0][index]
                j = indices[1][index]
                box = array.box if periodic else None
                distance = struc.distance(array[i], array[j], box=box)
            if len(indices) == 3:
                # multi_model = True -> AtomArrayStack
                m = indices[0][index]
                i = indices[1][index]
                j = indices[2][index]
                box = array.box[m] if periodic else None
                distance = struc.distance(array[m,i], array[m,j], box=box)
            try:
                assert distance == pytest.approx(CUTOFF, abs=1e-4)
            except AssertionError:
                print(f"Model {m}, Atoms {i} and {j}")
                raise
    
    stack = load_structure(join(data_dir("structure"), "1gya.mmtf"))
    stack.box = np.array([
        np.diag(np.max(coord, axis=0) - np.min(coord, axis=0) + 10)
        for coord in stack.coord
    ])
    stack.coord -= np.min(stack.coord, axis=-2)[:, np.newaxis, :] - 5
    if multi_model:
        array = stack
    else:
        array = stack[0]

    # Use adjacency matrices instead of pairwise distances
    # for compuational efficiency
    ref_matrix, ref_matrix_pbc = get_matrices(array)

    array = struc.translate(array, translation_vector)
    array.coord = struc.move_inside_box(array.coord, array.box)
    moved_matrix, moved_matrix_pbc = get_matrices(array)
    # The translation and the periodic move should not
    # alter PBC-aware pairwise distances
    assert_equal_matrices(array, ref_matrix_pbc, moved_matrix_pbc, True)
    # Non-PBC-aware distances should change,
    # otherwise the atoms do not go over the periodic boundary
    # and the test does not make sense
    with pytest.raises(AssertionError):
        assert_equal_matrices(array, ref_matrix, moved_matrix, False)

    array = struc.remove_pbc(array)
    restored_matrix, restored_matrix_pbc = get_matrices(array)
    # Both adjacency matrices should be equal to the original ones,
    # as the structure should be completely restored
    assert_equal_matrices(array, ref_matrix_pbc, restored_matrix_pbc, True)
    assert_equal_matrices(array, ref_matrix,     restored_matrix,     False)


@pytest.mark.parametrize("multi_model", [True, False])
def test_remove_pbc_selections(multi_model):
    """
    This test makes no assertions, it only test whether an exception
    occurs, when the `selection` parameter is given in `remove_pbc()`.
    """
    array = load_structure(join(data_dir("structure"), "3o5r.mmtf"))
    if multi_model:
        array = struc.stack([array, array])
    
    struc.remove_pbc(array)
    struc.remove_pbc(array, array.chain_id[0])
    struc.remove_pbc(array, struc.filter_amino_acids(array))
    struc.remove_pbc(array, [struc.filter_amino_acids(array),
                       (array.res_name == "FK5")])
    # Expect error when selectinf an atom multiple times
    with pytest.raises(ValueError):
        struc.remove_pbc(array, [struc.filter_amino_acids(array),
                                 (array.atom_name == "CA")])